# coding: utf-8
"""
    RankControl.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import os
import numpy as np
from bisect import bisect_right
from matplotlib.patches import Rectangle
from molass_legacy.Models.ElutionCurveModels import emg
import molass_legacy.KekLib.DebugPlot as plt

def demo():

    x = np.arange(300)

    fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(21,5))

    fig.suptitle("Rank Control Demo", fontsize=30)

    color_pairs = [('green', 'green'), ('red', 'green'), ('red', 'red')]
    for ax, cpair in zip(axes, color_pairs):
        ys = []
        tops = []
        ty = np.zeros(len(x))
        for h, mu in [(1, 130), (0.5, 170)]:
            y = emg(x, h, mu, 30, 0)
            ty += y
            ys.append(y)
            tops.append((mu, h))
        ax.plot(x, ty, color='orange', label='total elution')

        f = np.argmax(ty)
        t = f + bisect_right(-ty[f:], -ty[f]/2)

        ymin, ymax = ax.get_ylim()
        patch = Rectangle(
                (f, ymin),      # (x,y)
                t - f,          # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )

        ax.add_patch(patch)

        ranks = 0
        rankw = 0
        for k, y in enumerate(ys):
            ax.plot(x, y, ':', label='component %d' % k)
            px, py = tops[k]
            color = cpair[k]
            rankw += 1
            if color == "green":
                ranks += 1
            else:
                ranks += 2
            ax.plot(px, py, 'o', color=color)
        ax.set_title("Rank %d → Rank %d" % (ranks, rankw), fontsize=20)
        ax.legend(fontsize=16)

    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    plt.show()



# coding: utf-8
"""
    AtsasPlotter.py

    Copyright (c) 2016-2020, SAXS Team, KEK-PF
"""
from bisect import bisect_right
import numpy as np
from scipy import stats
# import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from matplotlib.font_manager import FontProperties
from DataUtils import get_in_folder

NUM_POINTS_REG  = 40

class AlmergePlotter:
    def __init__(self):
        self.fixed_font         = FontProperties()
        self.fixed_font.set_family( 'monospace' )

    def plot(self, datafiles, indeces, result, ex_data=None, aq_data=None, max_c=None, title=None, force_overlap_q=None, q_limit=None, descend=True, range_info=None, fig_info=None):

        if force_overlap_q is None:
            overlap_from_max = result.overlap_from_max
        else:
            file = datafiles[indeces[0]]
            array = np.loadtxt( file )
            x = array[:, 0]
            overlap_from_max = bisect_right(x, force_overlap_q)
            x = None

        print( 'forced overlap_from_max=', overlap_from_max )

        if fig_info is None:
            fig, axes_ = plt.subplots(nrows=2, ncols=2, figsize=(22, 12))
        else:
            fig, axes_ = fig_info

        ax0 = axes_[0,0]
        ax1 = axes_[0,1]
        ax2 = axes_[1,1]
        ax3 = axes_[1,0]
        axes = [ax1, ax2, ax3]

        side = ' (descend)' if descend else ' (ascend)'
        fig.suptitle("Comparison against ALMERGE using " + get_in_folder() + side, fontsize=30)

        if title is not None:
            ax1.set_title( title + ' - Input' )
            ax2.set_title( title + ' - Overlapped' )
            ax3.set_title( title + ' - Overlapped (Zoomed)' )

        if range_info is not None:
            ecurve, (f, t) = range_info
            ax0.plot(ecurve.x, ecurve.y, color='orange')
            ymin, ymax = ax0.get_ylim()
            ax0.set_ylim(ymin, ymax)
            p = Rectangle(
                    (f, ymin),      # (x,y)
                    t - f,          # width
                    ymax - ymin,    # height
                    facecolor   = 'cyan',
                    alpha       = 0.2,
                )
            ax0.add_patch( p )

        x = None
        plt_slice = None
        reg_slice = None
        x2 = None
        xr = None
        max_yp  = None
        max_i   = None
        max_k   = None
        min_yp  = None
        min_i   = None
        min_k   = None

        plt_list = []

        for k, i in enumerate(indeces):
            file = datafiles[i]
            array = np.loadtxt( file )
            print( i, file, array.shape,  )

            if x is None:
                x = array[:, 0]
                if q_limit is None:
                    j = len(x)
                    sa_view_q = x[-1]/4
                else:
                    j = bisect_right(x, q_limit)
                    sa_view_q = q_limit/4

                sa_view_xlim = sa_view_q**2
                plt_slice = slice(0, j)
                reg_end = min( j, overlap_from_max + NUM_POINTS_REG )
                reg_slice = slice( overlap_from_max, reg_end )
                x2 = x[plt_slice]**2
                xr  = x2[reg_slice]

            y   = array[plt_slice, 1]
            y_  = np.log(y)
            yr  = y_[reg_slice]
            slope, intercept, r_value, p_value, std_err = stats.linregress( xr, yr )
            yp = slope * xr[0] + intercept
            if max_yp is None or yp > max_yp:
                max_yp  = yp
                max_i   = i
                max_k   = k
            if min_yp is None or yp < min_yp:
                min_yp  = yp
                min_i   = i
                min_k   = k
            plt_list.append( [ y_, yp, slope, intercept ] )

        # print( 'min_i=', min_i, 'max_i=', max_i )

        for k, rec in enumerate(plt_list):
            y_, yp, slope, intercept = rec
            lin_y = slope * xr + intercept
            yp_shift = max_yp - yp

            i = indeces[k]
            if i == max_i:
                c = 'orange'
                alpha = 0.5
                label='max_c'
            elif i == min_i:
                c = 'green'
                alpha = 0.5
                label='min_c'
                min_y = y_ + yp_shift
            else:
                c = 'gray'
                alpha = 0.1
                label=None
            ax1.plot( x2, y_, color=c, alpha=alpha, label=label )
            ax1.plot( xr, lin_y, ':', color='red', alpha=0.5 )
            ax1.plot( xr[0], lin_y[0], 'o', color='red', markersize=5 )
            for ax in axes[1:]:
                ax.plot( x2, y_ + yp_shift, color=c, alpha=alpha, label=label )
                ax.plot( xr[0], lin_y[0] + yp_shift, 'o', color='red', markersize=5 )

        dev_slice = slice( 0, overlap_from_max )
        min_y_slice = min_y[dev_slice]
        zeros = np.zeros(overlap_from_max)

        if ex_data is None:
            ex_data = result.exz_array

        y_ex = np.log( ex_data[plt_slice,1] )
        diff = y_ex[dev_slice] - min_y_slice
        a_pos_dev = np.sum( np.max( [ zeros,  diff ], axis=0 ) )
        a_neg_dev = np.sum( np.max( [ zeros, -diff ], axis=0 ) )
        for ax in axes[1:]:
            ax.plot( x2, y_ex, color='red', label='almerge', alpha=0.5 )

        if aq_data is not None:
            scale = ex_data[overlap_from_max,1]/aq_data[overlap_from_max,1]
            # scale = max_c
            y_aq = np.log( aq_data[plt_slice,1]*scale )
            diff = y_aq[dev_slice] - min_y_slice
            k_pos_dev = np.sum( np.max( [ zeros,  diff ], axis=0 ) )
            k_neg_dev = np.sum( np.max( [ zeros, -diff ], axis=0 ) )
            for ax in axes[1:]:
                ax.plot( x2, y_aq, color='cyan', label='A(q)', alpha=0.5 )

        # plot again to be visible
        ax2.plot( xr[0], max_yp, 'o', color='red', markersize=5 )

        ymin, ymax = ax1.get_ylim()
        yoffset = ( ymax - ymin ) * 0.2

        ax1.annotate( 'Points to be made to overlap', xy=(xr[0], min_yp), fontsize=16,
                                xytext=( xr[0], min_yp - yoffset ),
                                ha='left',
                                arrowprops=dict( headwidth=3, width=0.3, color='gray' ),
                                )

        xmin, xmax = ax2.get_xlim()
        ymin, ymax = ax2.get_ylim()

        def annotate_overlap_point(ax, side):
            xmin, xmax = ax.get_xlim()
            ymin, ymax = ax.get_ylim()
            xmid = (xmax + xmin)/2
            if xr[0] < xmid:
                xoffset = ( xmax + xmin ) * 0.1
                yoffset = ( ymax - ymin ) * 0.05
                ha = 'left'
            else:
                xoffset = -( xmax + xmin ) * 0.1
                yoffset = ( ymax - ymin ) * 0.2
                ha = 'center'

            ax.annotate( 'Point made to overlap', xy=(xr[0], max_yp), fontsize=16,
                                    xytext=( xr[0] + xoffset, max_yp + yoffset ), ha=ha,
                                    arrowprops=dict( headwidth=3, width=0.3, color='gray' ),
                                    )

        xmin3 = xmin/8
        xmax3 = max(sa_view_xlim, xr[0]+abs(xmin3))
        ax3.set_xlim(xmin3, xmax3)
        w = (xmax3 - xmin)/(xmax - xmin)

        sa_view_ylim = min(ymin * w + ymax * (1 - w), max_yp)

        aslice = slice(0,overlap_from_max)
        ploted_max = max(np.max(y_ex[aslice]), np.max(y_aq[aslice]))
        ymax_ = ymax - (ymax - ploted_max)*3/4
        ax3.set_ylim( sa_view_ylim, ymax_ )

        for k, ax in enumerate(axes[1:]):
            annotate_overlap_point(ax, k)

        if False:
            if ex_data is not None and aq_data is not None:
                xmin, xmax = ax3.get_xlim()
                ymin, ymax = ax3.get_ylim()
                yoffset = ( ymax - ymin ) * 0.03

                tx = xmin * 0.75 + xmax * 0.25
                ty = ymin * 0.25 + ymax * 0.75
                ax3.text( tx, ty,           'almerge : pos_dev/neg_dev=%5.3f' % ( a_pos_dev/a_neg_dev ), fontproperties=self.fixed_font )
                ax3.text( tx, ty-yoffset,   'A(q)    : pos_dev/neg_dev=%5.3f' % ( k_pos_dev/k_neg_dev ), fontproperties=self.fixed_font )

        for ax in axes:
            ax.set_xlabel('Q²', fontsize=16)
            ax.set_ylabel('Ln(I)', fontsize=16)
            ax.legend(fontsize=16)

        fig.tight_layout()
        fig.subplots_adjust(top=0.92)
        # plt.show()

    def show(self):
        plt.show()



# coding: utf-8
"""

    DebugPlotUtils.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF

"""
from time import sleep
import matplotlib.pyplot as _plt
from molass_legacy.KekLib.OurTkinter import Tk
from molass_legacy.KekLib.TkUtils import adjusted_geometry
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.DebugPlot import set_plot_env

root = None

def mpl_setup(use_mpl=False):
    if use_mpl:
        ret_module = _plt
    else:
        global root
        root = Tk.Tk()
        root.geometry( adjusted_geometry( root.geometry() ) )
        root.withdraw()
        set_plot_env( root )
        ret_module = dplt

    return ret_module

def mpl_teardown(use_mpl, auto):
    if use_mpl:
        if auto:
            sleep(0.5)
    else:
        global root
        if auto:
            root.after( 500, root.quit )
        else:
            root.after( 0, root.quit )
        root.mainloop()
        root.destroy()



"""
    UV.WYF_Ratios.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from bisect import bisect_right

WAVELENGTHS = [280, 275, 258]

def get_wyf_indeces(wv):
    return [bisect_right(wv, wl) for wl in WAVELENGTHS]

def compute_ratios_from_Puv(wv, Puv, proportions=None, excl_prop=0.05):
    indecdes = get_wyf_indeces(wv)

    if proportions is None:
        colums = slice(None, None)
    else:
        colums = []
        for i, p in enumerate(proportions):
            if p > excl_prop:
                colums.append(i)

    ret_list = []
    for y in Puv.T[colums]:
        a280, a275, a258 = y[indecdes]
        r1 = a275/a280
        r2 = a258/a280
        ret_list.append((r1, r2))

    return ret_list

if __name__ == '__main__':
    pass



# coding: utf-8
"""
    SubtractorFrame.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import os
import numpy as np
from difflib import SequenceMatcher
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk, is_empty_val
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy.KekLib.TkUtils import is_low_resolution
try:
    import molass_legacy.KekLib.CustomMessageBox         as MessageBox
except:
    import OurMessageBox            as MessageBox
from XrayData import XrayData
from DataSubtractor import subtract_impl

TITLE_TEXTS = ["Sample Data", "Buffer Data", "Subtracted Data"]

class SubtractorFrame(Tk.Frame):
    def __init__(self, parent):
        self.busy = False
        self.data_list = [None]*3
        self.zlim_3d = None
        self.ylim_2d = None
        Tk.Frame.__init__(self, parent)
        self.body(self)

    def body(self, body_frame):
        tk_set_icon_portable(self)

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        self.current = 0
        self.frames = []

        frame_3d = Tk.Frame(body_frame)
        frame_3d.pack()
        self.frames.append(frame_3d)
        frame_2d = Tk.Frame(body_frame)
        frame_2d.pack()
        frame_2d.pack_forget()
        self.frames.append(frame_2d)

        cframe_3d = Tk.Frame(frame_3d)
        cframe_3d.pack()
        tbframe_3d = Tk.Frame(frame_3d)
        tbframe_3d.pack(fill=Tk.X, expand=1)

        cframe_2d = Tk.Frame(frame_2d)
        cframe_2d.pack()
        tbframe_2d = Tk.Frame(frame_2d)
        tbframe_2d.pack(fill=Tk.X, expand=1)

        folder_entry_width = 60
        num_files_label_width = 6

        grid_row = -1

        grid_row += 1
        label = Tk.Label(iframe, text="Folder Path")
        label.grid(row=grid_row, column=1)
        label = Tk.Label(iframe, text="File Name Pattern")
        label.grid(row=grid_row, column=2, padx=10)
        label = Tk.Label(iframe, text="Files")
        label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Sample Data")
        label.grid(row=grid_row, column=0)

        self.in_folder1 = Tk.StringVar()
        self.in_files1 = []
        self.name_pattern1 = Tk.StringVar()
        self.num_files1 = Tk.IntVar()
        self.in_folder1_cb = lambda: self.on_entry_in_folder(self.in_folder1, self.in_files1, self.name_pattern1, self.num_files1, 0)
        in_folder1_entry = FolderEntry(iframe, textvariable=self.in_folder1, width=folder_entry_width,
                                on_entry_cb=self.in_folder1_cb)
        in_folder1_entry.grid(row=grid_row, column=1)
        # in_folder1_entry.bind('<Leave>', lambda *args:self.in_folder1_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern1)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files1, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Buffer Data")
        label.grid(row=grid_row, column=0)
        self.in_folder2 = Tk.StringVar()
        self.in_files2 = []
        self.name_pattern2 = Tk.StringVar()
        self.num_files2 = Tk.IntVar()
        self.in_folder2_cb = lambda: self.on_entry_in_folder(self.in_folder2, self.in_files2, self.name_pattern2, self.num_files2, 1)
        in_folder2_entry = FolderEntry(iframe, textvariable=self.in_folder2, width=folder_entry_width,
                                on_entry_cb=self.in_folder2_cb)
        in_folder2_entry.grid(row=grid_row, column=1, pady=5)
        # in_folder2_entry.bind('<Leave>', lambda *args:self.in_folder2_cb())

        label = Tk.Label(iframe, textvariable=self.name_pattern2)
        label.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        num_files_label = Tk.Label(iframe, textvariable=self.num_files2, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        label = Tk.Label(iframe, text="Subtracted Data")
        label.grid(row=grid_row, column=0)
        self.out_folder = Tk.StringVar()
        out_folder_entry = FolderEntry(iframe, textvariable=self.out_folder, width=folder_entry_width,
                                on_entry_cb=self.on_entry_out_folder)
        out_folder_entry.grid(row=grid_row, column=1)
        out_folder_entry.bind('<Leave>', lambda *args: self.on_entry_out_folder())

        self.name_pattern3 = Tk.StringVar()
        entry = Tk.Entry(iframe, textvariable=self.name_pattern3, width=20)
        entry.grid(row=grid_row, column=2, padx=10, sticky=Tk.W)

        self.num_files3 = Tk.IntVar()
        num_files_label = Tk.Label(iframe, textvariable=self.num_files3, width=num_files_label_width, justify=Tk.RIGHT)
        num_files_label.grid(row=grid_row, column=3)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        self.subtract_btn = Tk.Button(iframe, text="Subtract", width=10, command=self.subtract, state=Tk.DISABLED)
        self.subtract_btn.grid(row=grid_row, column=2, columnspan=2, sticky=Tk.E)

        grid_row += 1
        space = Tk.Label(iframe)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="Progress")
        label.grid(row=grid_row, column=0)
        self.mpb = ttk.Progressbar(iframe,orient ="horizontal", length=600, mode="determinate")
        self.mpb.grid(row=grid_row, column=1, columnspan=3, padx=5)

        self.build_canvas_3d(cframe_3d, tbframe_3d)
        self.build_canvas_2d(cframe_2d, tbframe_2d)

    def build_canvas_3d(self, cframe, tbframe):
        figsize = (15,5) if is_low_resolution() else (21,7)
        fig, axes_3d = plt.subplots(nrows=1, ncols=3, figsize=figsize, subplot_kw=dict(projection='3d'))
        self.fig = fig
        self.axes_3d = axes_3d
        self.mpl_canvas_3d = FigureCanvasTkAgg( self.fig, cframe )
        self.mpl_canvas_3d_widget = self.mpl_canvas_3d.get_tk_widget()
        self.mpl_canvas_3d_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas_3d.mpl_connect( 'button_press_event', self.on_button_press )
        self.toolbar = NavigationToolbar(self.mpl_canvas_3d, tbframe)
        self.toolbar.update()
        fig.tight_layout()
        fig.subplots_adjust(top=0.92)
        self.mpl_canvas_3d.draw()

    def build_canvas_2d(self, cframe, tbframe):
        figsize = (15,5) if is_low_resolution() else (21,7)
        fig, axes_3d = plt.subplots(nrows=1, ncols=3, figsize=figsize)
        self.fig_2d = fig
        self.axes_2d = axes_3d
        self.mpl_canvas_2d = FigureCanvasTkAgg( self.fig_2d, cframe )
        self.mpl_canvas_2d_widget = self.mpl_canvas_2d.get_tk_widget()
        self.mpl_canvas_2d_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas_2d.mpl_connect( 'button_press_event', self.on_button_press )
        self.toolbar = NavigationToolbar(self.mpl_canvas_2d, tbframe)
        self.toolbar.update()
        fig.tight_layout()
        fig.subplots_adjust(top=0.92)
        self.mpl_canvas_2d.draw()

    def on_button_press(self, event):
        if event.dblclick:
            self.toggle_2d3d()

    def toggle_2d3d(self):
        previous = self.current
        self.current = 1 - self.current
        print('toggle_2d3d', previous, self.current)
        self.frames[previous].pack_forget()
        self.frames[self.current].pack()

    def on_entry_in_folder(self, in_folder, in_files, name_pattern, num_files, k):
        self.update()

        xray_data = XrayData(in_folder.get())
        self.data_list[k] = xray_data

        files = xray_data.files
        in_files.clear()
        in_files += files
        if len(files) > 0:
            dir_, file = os.path.split(files[0])
            name_pattern.set(file)
        num_files.set( len(files) )
        self.update()

        ax = self.axes_3d[k]
        ax.set_title(TITLE_TEXTS[k], fontsize=16)
        ec_color = 'orange' if k == 0 else 'gray'
        xray_data.plot(ax=ax, title=None, ec_color=ec_color)

        if self.zlim_3d is None:
            self.zlim_3d = ax.get_zlim()
        else:
            ax.set_zlim(self.zlim_3d)

        self.mpl_canvas_3d.draw()

        ax = self.axes_2d[k]
        ax.set_title(TITLE_TEXTS[k], fontsize=16)
        if xray_data.e_curve is None:
            x = xray_data.j
            k = xray_data.e_index
            y = xray_data.data[k,:]
        else:
            x = xray_data.e_curve.x
            y = xray_data.e_curve.y
        color = 'orange' if k == 0 else 'gray'
        ax.plot(x, y, color=color)

        if self.ylim_2d is None:
            self.ylim_2d = ax.get_ylim()
        else:
            ax.set_ylim(self.ylim_2d)

        self.mpl_canvas_2d.draw()

    def on_entry_out_folder(self):
        _, in_folder1 = os.path.split(self.in_folder1.get())
        name_pattern1 = self.name_pattern1.get()
        _, in_folder2 = os.path.split(self.in_folder2.get())
        name_pattern2 = self.name_pattern2.get()
        # print((in_folder1, name_pattern1), (in_folder2, name_pattern2))
        _, out_folder = os.path.split(self.out_folder.get())

        if is_empty_val(in_folder1) or is_empty_val(in_folder2) or is_empty_val(out_folder):
            return

        ret_pattern = self.generate_pattern(in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder)
        self.name_pattern3.set(ret_pattern)
        self.subtract_btn.config(state=Tk.NORMAL)

    def subtract(self):
        num_files1 = len(self.in_files1)
        num_files2 = len(self.in_files2)

        if num_files1 != num_files2:
            MessageBox.showerror("", "Numbers of files are different: %d != %d" % (num_files1, num_files2))
            return

        name_pattern = self.name_pattern3.get().replace('_00000', '_%05d')
        print('name_pattern=', name_pattern)

        self.busy = True
        process_option = 0
        out_folder = self.out_folder.get()
        ret_data_list = subtract_impl(self.in_files1, self.in_files2, process_option, out_folder, name_pattern, dialog=self)
        array = np.array(ret_data_list)
        q = array[0,:,0]
        data = array[:,:,1].T
        error = array[:,:,2].T
        xray_data = XrayData(out_folder, data=data, q=q, error=error)

        k = 2
        ax = self.axes_3d[k]
        ax.set_title(TITLE_TEXTS[k], fontsize=16)
        xray_data.plot(ax=ax, title=None, ec_color='brown')
        ax.set_zlim(self.zlim_3d)
        self.mpl_canvas_3d.draw()

        ax = self.axes_2d[k]
        ax.set_title(TITLE_TEXTS[k], fontsize=16)

        x = xray_data.e_curve.x
        y = xray_data.e_curve.y
        ax.plot(x, y, color='brown')

        ylim = list(ax.get_ylim())  # 'tuple' object does not support item assignment
        ylim[1] = self.ylim_2d[1]
        for ax in self.axes_2d:
            ax.set_ylim(tuple(ylim))

        self.busy = False

    def is_busy(self):
        return self.busy

    def generate_pattern(self, in_folder1, name_pattern1, in_folder2, name_pattern2, out_folder):
        debug = False
        match = SequenceMatcher(None, in_folder1, in_folder2).find_longest_match(0, len(in_folder1), 0, len(in_folder2))
        if debug:
            print(match)
            print(in_folder1[match.a: match.a + match.size])

        match = SequenceMatcher(None, name_pattern1, name_pattern2).find_longest_match(0, len(name_pattern1), 0, len(name_pattern2))
        common_pattern = name_pattern1[match.a: match.a + match.size]
        if debug:
            print(match)
            print(common_pattern)

        return "Diff" + common_pattern



# coding: utf-8
"""
    Rgg.RgRefiner.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.interpolate import UnivariateSpline
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import egh_pdf

TAU_LIMIT_RATIO = 0.5   # relative to sigma
VERY_SMALL_VALUE = 1e-10
XRAY_SCALE = 1

class RgRefiner:
    def __init__(self, init_params):
        self.init_params = np.array(init_params)

    def fit(self, x, y, rg, mask=None, seed=None, min_overlap=False, debug=False):
        spline = UnivariateSpline(x, y, s=0)
        num_components = len(self.init_params)
        if mask is None:
            xm = x
            ym = y
            rgm = rg
        else:
            xm = x[mask]
            ym = y[mask]
            rgm = rg[mask]  

        max_y = np.max(ym)
        xr_scale = XRAY_SCALE
        rg_scale = max_y/np.average(self.init_params[:,4])              # average rgs
        tau_scale = max_y/np.average(self.init_params[:,2])   # average sigmas
        overlap_base = max_y*0.01

        params_zeros = np.zeros(len(self.init_params))
        debug_internal = False

        plot_history = debug
        get_overlap = False
        if plot_history:
            history_data = []

        def obj_func(p):
            nonlocal debug_internal
            p_ = p.reshape(self.init_params.shape)
            ty = np.zeros(len(xm))
            cy_list = []
            for k in range(num_components):
                w, m, s, t = p_[k,0:4]
                cy = w * egh_pdf(xm, m, s, t)
                ty += cy
                cy_list.append(cy)

            scale = max_y/np.max(ty)
            trg = np.zeros(len(xm))
            for k in range(num_components):
                crg = p_[k,4]
                trg += crg*cy_list[k]/ty

            if debug_internal:
                plt.push()
                fig, ax = plt.subplots()
                axt = ax.twinx()
                axt.grid(False)

                ax.plot(xm, ym, color='C0')
                axt.plot(xm, rgm, color='C1')
                for cy in cy_list:
                    ax.plot(xm, scale*cy, ':')
                ax.plot(xm, scale*ty, ':', color='red')

                axt.set_ylim(0, 50)
                fig.tight_layout()
                debug_internal = plt.show()
                plt.pop()

            r1 = np.sum((scale*ty - ym)**2)
            r2 = np.sum((trg - rgm)**2)

            # should be 2*tau < sigma for each tau sigma
            tau_penalty = np.sum(np.max([params_zeros, np.abs(p_[:,3]) - p_[:,2]*TAU_LIMIT_RATIO], axis=0)**2)
            tau_penalty = max(VERY_SMALL_VALUE, tau_penalty)

            resid1 = np.log(xr_scale*r1)
            resid2 = np.log(rg_scale*r2)
            penalty = np.log(tau_scale*tau_penalty)

            if min_overlap:
                overlap = np.zeros(len(xm))
                for i in range(len(cy_list)-1):
                    cy1 = cy_list[i]
                    cy2 = cy_list[i+1]
                    overlap += np.abs(np.min([cy1, cy2], axis=0))       # np.abs() is intended to degrade negative elements
                if get_overlap:
                    return scale*overlap

                if True:
                    overlap_penalty = np.log(overlap_base + scale*np.average(overlap))  # overlap_base is intended to avoid overlap_penalty getting too small
                else:
                    old_settings = np.seterr(all='raise')
                    try:
                        overlap_penalty = np.log(overlap_base + scale*np.average(overlap))  # overlap_base is intended to avoid overlap_penalty getting too small
                    except Exception as exc:
                        print("overlap=", overlap)
                        print(exc)
                        debug_internal = True
                        obj_func(p)
                    np.seterr(**old_settings)
            else:
                overlap_penalty = 0

            if plot_history:
                history_data.append((resid1, resid2, penalty, overlap_penalty))

            return resid1 + resid2 + penalty + overlap_penalty

        bounds = []
        for w, m, s, t, rg in  self.init_params:
            bounds += [(0, 10), (0, m*2), (0, s*2), (-s, s), (0.5*rg, 2*rg)]

        # result = minimize(obj_func, self.init_params.flatten(), bounds=bounds)
        if seed is None:
            seed = np.random.randint(1000, 9999)
        # minimizer_kwargs = { "bounds":bounds }
        # minimizer_kwargs = { "method": "L-BFGS-B","bounds":bounds }
        minimizer_kwargs = None
        result = basinhopping(obj_func, self.init_params.flatten(), minimizer_kwargs=minimizer_kwargs, seed=seed)
        self.params = result.x.reshape(self.init_params.shape)
        self.func_value = result.fun
        self.seed = seed
        if min_overlap:
            get_overlap = True
            self.overlap = obj_func(result.x)
        else:
            self.overlap = None
        self.xm = xm
        self.ym = ym
        self.rgm = rgm

        if debug:
            print("sigma=", self.params[:,2])
            print("tau=", self.params[:,3])
            print("rg=", self.params[:,4])
            debug_internal = True
            obj_func(result.x)
            if plot_history:
                plt.push()
                fig, ax = plt.subplots()
                ax.plot(history_data)
                fig.tight_layout()
                plt.show()
                plt.pop()

    def get_components(self, x, y, max_y=None):
        if max_y is None:
            max_y = np.max(y)
        cy_list = []
        rg_list = []
        ty = np.zeros(len(x))
        for w, m, s, t, rg in self.params[:,0:5]:
            cy = w * egh_pdf(x, m, s, t)
            ty += cy
            cy_list.append(cy)
            rg_list.append(rg)
        scale = max_y/np.max(ty)
        return [scale*cy for cy in cy_list], scale*ty, rg_list

def plot_rgrefiner_result(ax, x, y, rgc, rr, mm_seed, min_overlap=False):
    cy_list, ty, rg_list = rr.get_components(x, y)

    ax.set_title("Refined Result with seeds=(%d, %d)" % (mm_seed, rr.seed), fontsize=16)
    ax.set_ylabel('Xray Intensity')
    ax.set_xlabel('Eno')
    ax.plot(x, y, color='orange')

    for k, cy in enumerate(cy_list):
        ax.plot(x, cy, ':', label='component-%d' % (k+1))
    ax.plot(x, ty, ':', color='red', label='total')
    ax.legend(loc='upper right')

    axt = ax.twinx()
    axt.grid(False)
    rg_ = np.zeros(len(x))
    for cy, rg in zip(cy_list, rg_list):
        rg_ += rg*cy/ty

    k = 0
    for slice_, state in zip(rgc.slices, rgc.states):
        if state == 0:
            continue

        x_, y_, rg = rgc.segments[k]
        label = "observed Rg" if k == 0 else None
        axt.plot(x_, rg, ':', color='C1', label=label)
        label = "reconstructed Rg" if k == 0 else None
        axt.plot(x_, rg_[slice_], color='gray', label=label)
        k += 1

    ymin, ymax = axt.get_ylim()
    axt.set_ylim(0, ymax)
    axt.legend(loc='upper left')

    if min_overlap:
        xm = rr.xm
        y1 = np.zeros(len(xm))
        y2 = rr.overlap
        ax.fill_between(xm, y1, y2, fc='pink', alpha=0.2)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    tx = xmin*0.95 + xmax*0.05
    ty = ymin*0.5 + ymax*0.5
    ax.text(tx, ty, "fv=%.4g" % rr.func_value, fontsize=20, alpha=0.3)

def spike_demo():
    from .SpikeData import generate_demo_data, make_availability_slices, FULL_PARAMS, normal_pdf
    from .GmmSpike import spike_demo_impl as gmm_impl

    x, y, rg, num_components = generate_demo_data()
    mm = gmm_impl([(x, y, rg, None)], num_components=num_components)

    weights = mm.weights_/np.sum(mm.weights_)
    init_params = [(w, m, np.sqrt(c), 0, r) for w, m, c, r in zip(weights, mm.means_[:,0], mm.covariances_[:,0,0], mm.means_[:,1])]
    init_params = np.array(sorted(init_params, key=lambda x: x[1]))
    print("init_params=", init_params)

    rr = RgRefiner(init_params)

    max_y = np.max(y)
    slices, states = make_availability_slices(y, max_y=max_y, min_ratio=0.03)
    mask = np.zeros(len(x), dtype=bool)
    for slice_, state in zip(slices, states):
        if state:
            mask[slice_] = True

    rr.fit(x, y, rg, mask, min_overlap=True, debug=True)



"""
    Peaks.PeakEditor.py

    Copyright (c) 2021-2025, SAXS Team, KEK-PF
"""
import copy
import logging
import queue
from bisect import bisect_right
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.KillableThread import Thread
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.Batch.FullBatch import FullBatch
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.QuickAnalysis.PeakUtils import guess_peak_params
from .ElutionModels import egh
from molass_legacy.Optimizer.FullOptInput import FullOptInput
from molass_legacy.Optimizer.FuncImporter import get_objective_function_info
from molass_legacy.Optimizer.OptConstants import MIN_NUM_PEAKS, MAX_NUM_PEAKS
from molass_legacy.Baseline.Constants import SLOPE_SCALE
from molass_legacy.Optimizer.FvScoreConverter import convert_score
from molass_legacy.Optimizer.OptimizerSettings import get_advanced_settings_text
from molass_legacy.RgProcess.RgCurve import ProgressCallback, draw_rg_bufer
from .PeProgressConstants import MAXNUM_STEPS, STOCH_INIT_STEPS, STARTED, RG_CURVE_OK, PREPARED
from molass_legacy._MOLASS.Version import is_developing_version

SKIP_RG_CURVE_CALCULATION = False
BETA_RELEASE = get_setting("beta_release")
ENABLE_SEPARATE_FOULING = get_setting("enable_separate_fouling")

if ENABLE_SEPARATE_FOULING:
    DRIFT_TYPES = ["Linear", "Linear + Uniform Fouling", "Linear + Separate Fouling"]
else:
    DRIFT_TYPES = ["Linear", "Linear + Fouling", "Disabled"]

def get_default_maxnum_trials(num_peaks):
    return max(30, num_peaks*10)

class PeakEditor(FullBatch, Dialog):
    def __init__(self, parent, sd, pre_recog, corrected_sd, treat, rg_folder=None, exact_num_peaks=None, strict_sec_penalty=False):
        FullBatch.__init__(self)
        self.logger = logging.getLogger(__name__)
        self.parent = parent

        if False:
            xr_curve1 = sd.get_xr_curve()
            xr_curve2 = corrected_sd.get_xr_curve()
            with dplt.Dp():
                fig, ax = dplt.subplots()
                ax.set_title("PeakEditor.__init__")
                ax.plot(xr_curve1.x, xr_curve1.y, label="raw data")
                ax.plot(xr_curve2.x, xr_curve2.y, label="corrected data")
                ax.legend()
                fig.tight_layout()
                dplt.show()

        self.sd = sd
        self.corrected_sd = corrected_sd
        self.pre_recog = pre_recog
        self.base_curve_info = treat.get_base_curve_info()
        uv_basemodel = get_setting("uv_basemodel")
        self.advanced = uv_basemodel == 1
        assert self.advanced
        
        self.dsets = None
        self.exact_num_peaks = exact_num_peaks
        self.strict_sec_penalty = strict_sec_penalty
        self.fullopt_class, self.class_code = None, None
        self.fullopt_input = FullOptInput(sd=sd, corrected_sd=corrected_sd, rg_folder=rg_folder)
        self.applied = False
        self.user_positions = None
        self.rg_line = None
        self.is_busy_ = True
        self.watching = False
        self.stop_watching = False
        self.title_text = None
        self.baselines = None
        self.baseline_params = None
        self.is_ready = False
        self.t0_upper_bound = None
        self.func_info = get_objective_function_info(self.logger)
        self.func_dict = self.func_info.func_dict
        self.key_list = self.func_info.key_list
        Dialog.__init__(self, parent, "Initial Estimator", visible=False)
    
    def is_busy(self):
        return self.is_busy_

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig)
        # print("ExtrapolSolverDialog: closed fig")
        Dialog.cancel(self)

    def show( self ):
        self.update()
        if not SKIP_RG_CURVE_CALCULATION:
            self.after(500, self.start_rg_curve_thread)
        self._show()

    def body(self, body_frame):
        upper_frame = Tk.Frame(body_frame)
        upper_frame.pack(fill=Tk.X)
        lower_frame = Tk.Frame(body_frame)
        lower_frame.pack(fill=Tk.X)

        lower_left = Tk.Frame(lower_frame)
        lower_left.pack(side=Tk.LEFT, padx=10)
        lower_right = Tk.Frame(lower_frame)
        lower_right.pack(side=Tk.LEFT, padx=10)

        if is_developing_version():
            devel_frame = Tk.Frame(lower_frame)
            devel_frame.pack(side=Tk.LEFT, padx=10)
            devel_button = Tk.Button(devel_frame, text="Devel Test", command=self.devel_test)
            devel_button.pack()

        lower_left1 = Tk.Frame(lower_left)
        lower_left1.pack(fill=Tk.X)
        tool_frame = Tk.Frame(lower_left1)
        tool_frame.pack(side=Tk.LEFT)
        lower_left2 = Tk.Frame(lower_left)
        lower_left2.pack()

        fig, axes = plt.subplots(ncols=3, figsize=(18,5))
        self.fig = fig
        self.axes = axes
        ax2 = axes[1]
        self.axt = ax2.twinx()
        self.axt.grid(False)
        self.mpl_canvas = FigureCanvasTkAgg(fig, upper_frame)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        uv_x, uv_y, xr_x, xr_y, baselines = self.get_curve_xy(return_baselines=True)
        uv_y_ = uv_y - baselines[0]
        xr_y_ = xr_y - baselines[1]
        uv_peaks, xr_peaks = self.get_modeled_peaks(uv_x, uv_y_, xr_x, xr_y_)
        self.set_lrf_src_args1(uv_x, uv_y, xr_x, xr_y, baselines)
        self.xr_peaks_orig = copy.deepcopy(xr_peaks)

        self.draw_elements(uv_peaks, xr_peaks, uv_x, uv_y, xr_x, xr_y, baselines=baselines)
        self.build_lower(lower_left2, lower_right)
        self.draw_title()       # must be done after build_lower

        self.toolbar = NavigationToolbar(self.mpl_canvas, tool_frame)
        self.toolbar.update()

        self.popup_menu = None
        self.mpl_canvas.mpl_connect('button_press_event', self.on_figure_click)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack(fill=Tk.X, padx=20)

        num_buttons = 8
        for j in range(num_buttons):
            box.columnconfigure(j, weight=1)

        w = Tk.Button(box, text="◀ Cancel", width=10, command=self.user_cancel)
        col = 0
        w.grid(row=0, column=col, pady=10)
        self.cancel_btn = w

        self.buttons_to_update = []

        col += 1
        w = Tk.Button(box, text="▽ Bounds Inspection", width=18, command=self.show_bounds, state=Tk.DISABLED)
        w.grid(row=0, column=col, pady=10)
        self.bounds_btn = w
        self.buttons_to_update.append(w)

        col += 1
        w = Tk.Button(box, text="▽ Show Parameters", width=18, command=self.show_params, state=Tk.DISABLED)
        w.grid(row=0, column=col, pady=10)
        self.params_btn = w
        self.buttons_to_update.append(w)

        if is_developing_version():
            col += 1
            w = Tk.Button(box, text="▽ Redraw Scores", width=18, command=self.redraw_scores, state=Tk.DISABLED)
            w.grid(row=0, column=col, pady=10)
            self.redraw_btn = w
            self.buttons_to_update.append(w)

        col += 1
        w = Tk.Button(box, text="▽ Complementary View", width=20, command=self.show_complementary_view, state=Tk.DISABLED)
        w.grid(row=0, column=col, pady=10)
        self.opposite_btn = w
        self.buttons_to_update.append(w)

        col += 1
        w = Tk.Button(box, text="▽ Edit Canvas", width=18, command=self.show_editcanvas, state=Tk.DISABLED)
        w.grid(row=0, column=col, pady=10)
        self.editcanvas_btn = w
        self.buttons_to_update.append(w)

        col += 1
        w = Tk.Button(box, text="▽ CPD Decompose", width=18, command=self.show_cpd_decompose, state=Tk.DISABLED)
        w.grid(row=0, column=col, pady=10)
        self.editcanvas_btn = w
        self.buttons_to_update.append(w)

        col += 1
        self.optimize_btn_blink = BlinkingFrame(box)
        self.optimize_btn_blink.grid(row=0, column=col, pady=10)
        w = Tk.Button(self.optimize_btn_blink, text="▶ Optimize", width=14, command=self.ok, state=Tk.DISABLED)
        w.pack()
        self.optimize_btn_blink.objects = [w]
        self.optimize_btn = w
        self.buttons_to_update.append(w)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def user_cancel(self, ask=True):
        test_pattern = get_setting("test_pattern")
        if test_pattern is None and ask:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            ret = MessageBox.askyesno("Cancel Confirmation",
                "Are you sure to cancel?",
                parent=self)
            if not ret:
                return

        if self.watching:
            self.stop_watching = True
        else:
            self.cancel()

    def apply(self):
        self.applied = True

    def draw_title(self):
        from molass_legacy.Optimizer.OptimizerUtils import get_model_name, get_method_name
        fullopt_class, class_code = self.get_function_class()
        text = "Elution Decomposition of %s with model=%s method=%s" % (get_in_folder(), get_model_name(class_code), get_method_name())
        if self.title_text is None:
            self.title_text = self.fig.suptitle(text, fontsize=20)
        else:
            self.title_text.set_text(text)

    def draw_elements(self, uv_peaks, xr_peaks, uv_x, uv_y, xr_x, xr_y, baselines=None, draw_canvas=True):
        if False:
            import QuickAnalysis.ModeledPeaks
            from importlib import reload
            reload(QuickAnalysis.ModeledPeaks)
        from molass_legacy.QuickAnalysis.ModeledPeaks import plot_curve

        fig = self.fig
        axes = self.axes

        for ax in axes:
            ax.cla()

        ax1, ax2, ax3 = axes
        ax1.set_title("UV Elution at $\lambda$=%.3g" % get_setting("absorbance_picking"), fontsize=16)
        ax2.set_title("Xray Elution at Q=%.3g" % get_setting("intensity_picking"), fontsize=16)
        ax3.set_title("Objective Function Scores", fontsize=16)

        if baselines is None:
            uv_baseline, xr_baseline = None, None
        else:
            uv_baseline, xr_baseline = baselines
        self.uv_ty = plot_curve(ax1, uv_x, uv_y, uv_peaks, color='blue', baseline=uv_baseline)
        plot_curve(ax2, xr_x, xr_y, xr_peaks, color='orange', baseline=xr_baseline)
        self.xr_draw_info = [xr_x, xr_y, xr_peaks, xr_baseline]

        ax3.text(0.5, 0.5, "Not Ready", color="gray", ha="center", va="center", fontsize=20)

        fig.tight_layout()
        fig.subplots_adjust(top=0.8)
        if draw_canvas:
            self.mpl_canvas.draw()

    def build_lower(self, lower_left, lower_right):

        self.pbar = ttk.Progressbar(lower_left, orient ="horizontal", length=700, mode="determinate")
        self.pbar["maximum"] = MAXNUM_STEPS
        self.pbar.pack(pady=10)

        self.status_msg = Tk.StringVar()
        label = Tk.Label(lower_left, textvariable=self.status_msg, bg='white')
        label.pack(fill=Tk.X, pady=10)
        self.status_bar = label

        self.build_panel(lower_right)    # 

    def build_panel(self, frame):
        pframe = Tk.Frame(frame)
        pframe.pack(padx=50)

        temp_frame = Tk.Frame(pframe)
        temp_frame.grid(row=0, column=0, padx=5, pady=5)

        label = Tk.Label(temp_frame, text="Number of Components: ")
        label.grid(row=0, column=0, sticky=Tk.W)

        self.user_num_peaks = Tk.IntVar()
        num_peaks=len(self.peak_params_set[1])
        self.user_num_peaks.set(num_peaks)

        sbox  = Tk.Spinbox(temp_frame, textvariable=self.user_num_peaks,
                          from_=MIN_NUM_PEAKS, to=MAX_NUM_PEAKS, increment=1,
                          justify=Tk.CENTER, width=5,
                          state=Tk.DISABLED)
        sbox.grid(row=0, column=1)

        temp_frame = Tk.Frame(pframe)
        temp_frame.grid(row=1, column=0, padx=5, pady=5)

        label = Tk.Label(temp_frame, text="Baseline: ")
        label.grid(row=0, column=0, sticky=Tk.W)

        unified_baseline_type = get_setting("unified_baseline_type")
        self.drift_type = Tk.StringVar()
        self.drift_type.set(DRIFT_TYPES[unified_baseline_type - 1])
        self.drift_type_box = ttk.Combobox(master=temp_frame,
                    values=DRIFT_TYPES, textvariable=self.drift_type, width=15, state=Tk.DISABLED)
        self.drift_type_box.grid(row=0, column=1)

        temp_frame = Tk.Frame(pframe)
        temp_frame.grid(row=0, column=1, padx=5, pady=5)

        label = Tk.Label(temp_frame, text="Number of Iterations: ")
        label.grid(row=0, column=0, sticky=Tk.W)

        self.num_iter = Tk.IntVar()
        self.num_iter.set(20)
        self.num_iter_entry = Tk.Entry(temp_frame, textvariable=self.num_iter, width=5, justify=Tk.CENTER)
        self.num_iter_entry.grid(row=0, column=1)

        temp_frame = Tk.Frame(pframe)
        temp_frame.grid(row=1, column=1, padx=5, pady=5)

        label = Tk.Label(temp_frame, text="Random Number Seed: ")
        label.grid(row=0, column=0, sticky=Tk.W)

        self.seed = Tk.IntVar()
        self.seed.set(np.random.randint(100000, 999999))
        self.seed_entry = Tk.Entry(temp_frame, textvariable=self.seed, width=8, justify=Tk.CENTER)
        self.seed_entry.grid(row=0, column=1)

        self.param_init_type = Tk.IntVar()
        self.param_init_type.set(1)     # Known best

        temp_frame = Tk.Frame(pframe)
        temp_frame.grid(row=0, column=2, padx=5, pady=5)

        row = 0
        label = Tk.Label(temp_frame, text="Max Number of Trials: ")
        label.grid(row=row, column=0, sticky=Tk.E)

        self.maxnum_trials = Tk.IntVar()
        mt_init = get_default_maxnum_trials(num_peaks)
        self.maxnum_trials.set(mt_init)
        self.mn_sbox = Tk.Spinbox(temp_frame, textvariable=self.maxnum_trials,
                          from_=1, to=100, increment=1,
                          justify=Tk.CENTER, width=5)
        self.mn_sbox.grid(row=row, column=1, sticky=Tk.W)

        label_text = get_advanced_settings_text()
        if label_text > "":
            label = Tk.Label(pframe, text=label_text)
            label.grid(row=2, column=0, columnspan=3)

    def start_rg_curve_thread(self):
        self.queue = queue.Queue()
        self.rg_curve_thread = Thread(target=self.prepare_rg_curve, args=[self.queue], name="Rg Curve Thread")
        self.rg_curve_thread.start()
        self.watching = True
        self.after(500, self.watch_rg_curve_thread)

    def prepare_rg_curve(self, queue):
        queue.put([STARTED, None])

        if self.elution_model in [0, 2]:
            rg_curve_ok = RG_CURVE_OK
        else:
            rg_curve_ok = RG_CURVE_OK - STOCH_INIT_STEPS

        progress_cb = ProgressCallback(queue, STARTED, rg_curve_ok)
        self.dsets = self.fullopt_input.get_dsets(progress_cb=progress_cb, compute_rg=True, possibly_relocated=False)
        queue.put([rg_curve_ok, None])
        queue.put([PREPARED, None])

    def watch_rg_curve_thread(self):
        try:
            progress, p_info = self.queue.get()
            self.pbar["value"] = progress

            if p_info is not None and type(p_info) == tuple:
                xr_curve = self.ecurves[1]
                drawn = draw_rg_bufer(self.axt, p_info, self, xr_curve.x)   # this updates self.rg_line
                if drawn:
                    j = p_info[1]
                    self.update_status_bar("Computing Rg values near the %d-th elution." % j)
                    self.mpl_canvas.draw()

            elif progress < 0:
                self.rg_curve_thread.join()
                self.watching = False
                progress = MAXNUM_STEPS
                if self.elution_model == 1:
                    progress -= STOCH_INIT_STEPS
                self.pbar["value"] = progress
                self.update_status_bar("Rg curve is ready.")
                self.rg_curve_thread = None
                self.get_ready_for_scores_etc()
                # self.save_rg_buffer()
                return

        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "watch_prepare_thread")

        if self.stop_watching:
            self.rg_curve_thread.terminate()
            self.watching = False
            self.is_busy_ = False
            self.cancel()
        else:
            self.after(100, self.watch_rg_curve_thread)      # 100 must be shorter than Rg computation

    def update_status_bar(self, msg):
        self.logger.info(msg)
        self.status_msg.set(msg)
        self.update()

    def get_ready_for_scores_etc(self):
        # for the time being until None baseline is available
        self.after(1000, self.get_ready_for_optimization)

    def update_button_states(self):
        for k, widget in enumerate(self.buttons_to_update):
            if BETA_RELEASE and k in [0]:
                pass
            else:
                widget.config(state=Tk.NORMAL)

    def save_rg_buffer(self):
        rc_curve = self.dsets[1]
        rc_curve.save_buffer("rg_buffer.dat")

    def re_construct_optimizer(self):
        self.construct_optimizer()
        init_params = self.compute_init_params()
        self.fullopt.prepare_for_optimization(init_params)

    def draw_scores(self, init_params=None, draw_rg_curve=True, create_new_optimizer=True):

        if create_new_optimizer:
            self.construct_optimizer()
        else:
            # as in on-the-fly debug from self.devel_test
            pass

        if init_params is None:
            init_params = self.compute_init_params()

        self.fullopt.prepare_for_optimization(init_params)

        for ax in self.axes:
            ax.cla()
        self.axt.remove()

        ax1, ax2 = self.axes[0:2]
        axt = ax2.twinx()
        axt.grid(False)
        self.axt = axt

        axis_info = (self.fig, (*self.axes, axt))       # i.e., draw all
        ax1.set_title("UV Decomposition", fontsize=16)
        ax2.set_title("Xray Decomposition", fontsize=16)

        try:
            fv = self.fullopt.objective_func(self.fullopt.init_params, plot=True, axis_info=axis_info)
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "draw_scores: ", n=10)
            fv = np.inf

        ax3 = self.axes[2]
        ax3.set_title("Objective Function Scores in SV=%.3g" % convert_score(fv), fontsize=16)
        self.fig.tight_layout()
        self.mpl_canvas.draw()
        self.is_busy_ = False
        # self.optimize_btn_blink.start()

    def get_ready_for_optimization(self):
        self.draw_scores()      # includes self.construct_optimizer()
        self.update_button_states()
        self.is_ready = True

    def get_optimizer(self):
        return self.fullopt

    def get_init_params(self):
        return self.fullopt.init_params     # this includes extended params if self.fullopt supports them

    def clear_pframe(self):
        assert False

    def get_drift_type(self):
        drift_type = self.drift_type.get()
        if drift_type == DRIFT_TYPES[0]:
            return 'linear'
        elif drift_type == DRIFT_TYPES[1]:
            return 'integral'
        elif drift_type == DRIFT_TYPES[2]:
            return 'integral'
        else:
            assert False

    def on_figure_click(self, event):
        if event.button == 3 and self.is_ready:
            if event.inaxes in [self.axes[2]]:
                self.show_popup_menu(event)
                return

        if event.xdata is None:
            return

        if False:
            print("double click")
            i = None
            for k, ax in enumerate(self.axes):
                if ax == event.inaxes:
                    i = k
                    break

            print([i], "axis")
            if i == 0:
                a, b = self.peak_params_set[2:4]
                xr_x = (event.xdata - b)/a
            else:
                xr_x = event.xdata

            print("event.xdata=", event.xdata, "xr_x=", xr_x)
            self.try_insert_user_peak(xr_x)

    def show_popup_menu(self, event):
        from molass_legacy.KekLib.TkUtils import split_geometry

        self.create_popup_menu(event)
        canvas = self.mpl_canvas_widget
        cx = canvas.winfo_rootx()
        cy = canvas.winfo_rooty()
        w, h, x, y = split_geometry(canvas.winfo_geometry())
        self.popup_menu.post(cx + int(event.x), cy + h - int(event.y))

    def create_popup_menu(self, event):
        if self.popup_menu is None:
            self.popup_menu = Tk.Menu(self, tearoff=0 )
            self.popup_menu.add_command(label='Show Score Details', command=self.show_score_details)

    def try_insert_user_peak(self, xr_x):
        # to be moved to EditCanvas.py

        if self.user_positions is None:
            self.user_positions = np.array([params[1]  for params in self.peak_params_set[1]])

        xr_curve = self.ecurves[1]
        x = xr_curve.x
        min_dist = np.min(np.abs(self.user_positions - xr_x))
        ratio = min_dist/len(x)
        print("min_dist=", min_dist, "ratio=", ratio)
        if ratio > 0.05:
            # self.insert_user_peak(xr_x)
            assert False
        else:
            print("user_positions=", self.user_positions)

    def get_dsets(self):
        if self.dsets is None:
            sd = self.sd
            D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
            U, wlvec, uv_curve = sd.conc_array, sd.lvector, sd.get_uv_curve()
            dsets = ((xr_curve, D), None, (uv_curve, U))
            return dsets
        else:
            return self.dsets

    def get_n_iterations(self):
        return self.num_iter.get()

    def get_seed(self):
        return self.seed.get()

    def get_param_init_type(self):
        return self.param_init_type.get()

    def get_maxnum_trials(self):
        return self.maxnum_trials.get()

    def show_bounds(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.BoundsInspection
            reload(molass_legacy.Optimizer.BoundsInspection)
        from molass_legacy.Optimizer.BoundsInspection import BoundsInspection
        dialog = BoundsInspection(self.parent, self.fullopt)
        dialog.show()

    def show_params(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.ParamsInspection
            reload(molass_legacy.Optimizer.ParamsInspection)
        from molass_legacy.Optimizer.ParamsInspection import ParamsInspection
        if debug and is_developing_version():
            self.re_construct_optimizer()

        init_params = self.get_init_params()
        dialog = ParamsInspection(self.parent, init_params, self.dsets, self.fullopt)
        dialog.show()

    def redraw_scores(self):
        # nothing else required?
        self.draw_scores()

    def show_complementary_view(self, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.Optimizer.ComplementaryView
            reload(molass_legacy.Optimizer.ComplementaryView)
        from molass_legacy.Optimizer.ComplementaryView import ComplementaryView
        from molass_legacy.DataStructure.SvdDenoise import get_denoised_data

        dsets = self.get_dsets()
        work_folder = get_setting("analysis_folder")
        init_params = self.get_init_params()
        ov = ComplementaryView(self.parent, self.fullopt, -1, init_params, work_folder)
        ov.show()

    def show_editcanvas(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.Peaks.EditCanvas
            reload(molass_legacy.Peaks.EditCanvas)
        from .EditCanvas import EditCanvas
        xr_x, xr_y, xr_peaks, xr_baseline = self.xr_draw_info
        original_info = [xr_x, xr_y, self.xr_peaks_orig, xr_baseline]
        ec = EditCanvas(self, self.sd, self.ecurves, original_info, working_info=self.xr_draw_info)
        ec.show()
        if ec.applied:
            new_pps = ec.get_peak_params_set()
            self.redraw_with_new_params(new_pps)

    def show_cpd_decompose(self, debug=True):
        if debug:
            from importlib import reload
            import molass_legacy.GuinierTools.CpdDecompDirect
            reload(molass_legacy.GuinierTools.CpdDecompDirect)
        from molass_legacy.GuinierTools.CpdDecompDirect import cpd_direct_impl
        cpd_direct_impl(self)

    def redraw_with_new_params(self, new_pps):        
        self.peak_params_set = new_pps
        uv_curve, xr_curve = self.ecurves
        xr_x = xr_curve.x
        xr_y = xr_curve.y
        uv_x = uv_curve.x
        uv_y = uv_curve.y
        new_uv_peaks, new_xr_peaks = new_pps[0:2]
        self.draw_elements(new_uv_peaks, new_xr_peaks, uv_x, uv_y, xr_x, xr_y, baselines=self.baselines, draw_canvas=False)
        # recompute init_params
        # task:
        #   it would be better to code this update explicitly
        #   rather than specifing the intension with arguments as here below
        self.draw_scores(create_new_optimizer=True, init_params=None)

    def save_the_figure(self, file):
        self.fig.savefig(file)

    def show_score_details(self, debug=True):
        if debug:
            import molass_legacy.Optimizer.FvScoreDetails as details_module
            from importlib import reload
            reload(details_module)
        from molass_legacy.Optimizer.FvScoreDetails import FvScoreDetails
        if debug and is_developing_version():
            self.re_construct_optimizer()

        details = FvScoreDetails(self.parent, self.fullopt, self.init_params)
        details.show()

    def devel_test(self):
        from importlib import reload
        import molass_legacy.Peaks.PeakDevel
        reload(molass_legacy.Peaks.PeakDevel)
        from molass_legacy.Peaks.PeakDevel import devel_test_impl

        devel_test_impl(self)



# coding: utf-8
"""

    ファイル名：   CanvasDialog.py

    処理内容：

        デバッグ用の plot 

    Copyright (c) 2017-2018, Masatsuyo Takahashi, KEK-PF

"""
import matplotlib.pyplot                as plt
from matplotlib.backends.backend_tkagg  import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter             import Tk, Dialog
from molass_legacy.KekLib.BasicUtils             import get_caller_module
from molass_legacy.KekLib.TkSupplements          import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib          import NavigationToolbar
from molass_legacy.KekLib.TkUtils                import adjusted_geometry
import matplotlib

class CanvasDialog( Dialog ):
    def __init__( self, title, parent=None, adjust_geometry=False, figure=None ):

        if parent is None:
            parent = Tk.Toplevel()
            if adjust_geometry:
                parent.geometry( adjusted_geometry( parent.geometry() ) )
            parent.withdraw()
            self.created_parent = True
        else:
            self.created_parent = False

        self.grab = 'local'     # used in grab_set
        self.parent     = parent
        self.title_     = title
        self.applied    = None
        self.caller_module = get_caller_module( level=2 )
        self.figure     = figure
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'

    def show( self, draw_func, figsize=None, message=None, button_labels=[ "OK", "Cancel" ],
                toolbar=False, parent_arg=False ):
        self.draw_func  = draw_func
        self.figsize    = figsize
        self.message    = message
        assert( len(button_labels) == 2 )
        self.button_labels  = button_labels
        self.toolbar    = toolbar
        self.parent_arg = parent_arg

        Dialog.__init__( self, self.parent, self.title_ )
        # TODO: self.resizable(width=False, height=False)

    def destroy_parent( self ):
        if self.created_parent:
            self.parent.destroy()

    def body( self, body_frame ):   # overrides parent class method
        tk_set_icon_portable( self, module=self.caller_module )

        if self.message is not None:
            self.msg = Tk.Label( body_frame, text=self.message, bg='white' )
            self.msg.pack( fill=Tk.BOTH, expand=1, pady=20 )
            # msg.insert( Tk.INSERT, self.message )
            # msg.config( state=Tk.DISABLED )

        cframe = Tk.Frame( body_frame )
        cframe.pack()

        if self.figure is None:
            figsize_ = ( 18, 9 ) if self.figsize is None else self.figsize
            fig = plt.figure( figsize=figsize_ )
        else:
            fig = self.figure

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )

        # it seems that draw_func should be called after the creation of mpl_canvas
        # in order to enable 3d-rotation of the figure
        if self.parent_arg:
            self.draw_func( fig, self.parent )
        else:
            self.draw_func( fig )
        self.parent.config( cursor='' )

        if self.toolbar:
            self.toolbar = NavigationToolbar( self.mpl_canvas, cframe )
            self.toolbar.update()

        self.protocol( "WM_DELETE_WINDOW", self.ok )

    def buttonbox(self):
        '''add standard button box.

        override if you do not want the standard buttons
        '''

        box = Tk.Frame(self)

        w = Tk.Button(box, text=self.button_labels[0], width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.ok_button = w
        w = Tk.Button(box, text=self.button_labels[1], width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        self.cancel_button = w

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        box.pack()

    def ok( self, *args ):
        self.destroy_parent()
        self.applied  = True
        self.destroy()

    def cancel( self, *arg ):
        self.destroy_parent()
        self.applied  = False
        self.destroy()




"""
    RdfDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.animation import FuncAnimation
from bisect import bisect_right
from Theory.Web.ACS.fcc_coord_gen import generate_fcc_coords
from Theory.Web.ACS.analytic_rdf import RDF_AnalyticNorm
from .PyChemLib3D import BOX_LIMITS

def generate_sphere(r=4):
    particles = generate_fcc_coords(blurr=False)

    x = particles[:,0]
    y = particles[:,1]
    z = particles[:,2]
    sphere = x**2 + y**2 + z**2 < r**2
    return particles[sphere,:]

def make_shifted(displ, particles):
    ret_particles = particles.copy()
    for j in range(3):
        ret_particles[:,j] += displ[j]
    return ret_particles

def genarate_params(n):
    radii = np.random.normal(15, 1, n)
    print("radii=", radii)
    displs = np.random.uniform(-1,1,(n,3))
    print("displs=", displs)

    centers = [np.zeros(3)]
    for k in range(n):
        dx, dy, dz = displs[k,:]
        r = np.sqrt(dx**2 + dy**2 + dz**2)
        ratio = radii[k]/r
        centers.append(displs[k,:]*ratio)

    return radii, displs, centers

def generate_molecules(particles1, n, min_dist=9):
    while True:
        radii, displs, centers = genarate_params(n)
        exists_near_pair = False
        for i in range(len(centers)):
            for j in range(i+1, len(centers)):
                dist = np.sqrt(np.sum((centers[i] - centers[j])**2))
                if dist < min_dist:
                    exists_near_pair = True
                    break
        if exists_near_pair:
            print("exists_near_pair continue", dist)
            continue
        else:
            print("no near_pairs. ok")
            break

    molecules = [particles1]
    for k in range(n):
        dx, dy, dz = displs[k,:]
        r = np.sqrt(dx**2 + dy**2 + dz**2)
        ratio = radii[k]/r
        molecule = make_shifted(displs[k,:]*ratio, particles1)
        molecules.append(molecule)

    return molecules

"""
radii= [15.00525968 15.49340332 12.27811157 15.75436373]
displs= [[-0.97795202 -0.58386681  0.92010054]
 [ 0.32175131 -0.85231832 -0.46558907]
 [ 0.60603154  0.46890646  0.31046006]
 [ 0.49491593  0.63570313 -0.31370224]]
"""

def demo(center_pos_file=None, rdf_files=None):
    particles1 = generate_sphere()
    print(particles1.shape)

    if center_pos_file is None:
        num_molecules = 8
        num_shifts = num_molecules - 1
        molecules = generate_molecules(particles1, num_shifts)
    else:
        centers = np.loadtxt(center_pos_file)
        molecules = []
        for center in centers:
            molecule = make_shifted(center, particles1)
            molecules.append(molecule)
        num_molecules = len(molecules)

    all_particles = np.concatenate(molecules)
    print("all_particles.shape=", all_particles.shape)
    if rdf_files is None:
        r = np.linspace(0, 60, 200)
        gr_one = RDF_AnalyticNorm(molecules[0], r, 0.05)
        gr_all = RDF_AnalyticNorm(all_particles, r, 0.05)
        np.savetxt("rdf_one.dat", np.array([r, gr_one]).T)
        np.savetxt("rdf_all.dat", np.array([r, gr_all]).T)
    else:
        rdf_one = np.loadtxt(rdf_files[0])
        rdf_all = np.loadtxt(rdf_files[1])
        assert rdf_one.shape == rdf_all.shape
        r = rdf_one[:,0]
        gr_one = rdf_one[:,1]
        gr_all = rdf_all[:,1]

    gs = GridSpec(3,2)
    fig = plt.figure(figsize=(16,8))
    fig.suptitle("Intera/inter-particle Radial Distribution Demo for N=%d" % num_molecules, fontsize=20)

    ax1 = fig.add_subplot(gs[:,0], projection="3d")
    ax2 = fig.add_subplot(gs[0,1])
    ax3 = fig.add_subplot(gs[1,1])
    ax4 = fig.add_subplot(gs[2,1])

    ax1.set_title("3D View", fontsize=16)

    for particles in molecules:
        ax1.plot(*particles.T, "o", markersize=3, alpha=0.5)

    ax1.set_xlim(*BOX_LIMITS[0])
    ax1.set_ylim(*BOX_LIMITS[1])
    ax1.set_zlim(*BOX_LIMITS[2])

    k = bisect_right(r, 50)
    plot_slice = slice(0, k)
    r_ = r[plot_slice]

    ax2.set_title("RDF of one molecule", fontsize=16)
    ax2.plot(r_, gr_one[plot_slice])
    ax3.set_title("RDF of all molecules", fontsize=16)
    ax3.plot(r_, gr_all[plot_slice])
    ax4.set_title("RDF of all molecules (zoomed)", fontsize=16)
    ax4.plot(r_, gr_all[plot_slice])
    ymin, ymax = ax4.get_ylim()
    ax4.set_ylim(-0.1, 2)

    fig.tight_layout()
    plt.show()

def demo_lj():
    num_molecules = 10
    num_shifts = num_molecules - 1

    radii = np.random.normal(15, 1, num_shifts)
    print("radii=", radii)
    displs = np.random.uniform(-1,1,(num_shifts,3))
    print("displs=", displs)

    points = [(0.0, 0.0, 0.0)]

    for k in range(num_shifts):
        dx, dy, dz = displs[k,:]
        r = np.sqrt(dx**2 + dy**2 + dz**2)
        ratio = radii[k]/r
        points.append(displs[k,:]*ratio)

    points = np.array(points)

    fig, ax = plt.subplots(subplot_kw=dict(projection="3d"), figsize=(8,8))
    point_arts, = ax.plot(*points.T, "o")
    ax.set_xlim(-20, 20)
    ax.set_ylim(-20, 20)
    ax.set_zlim(-20, 20)
    fig.tight_layout()

    pos_list = []
    points_ = points.copy()
    for n in range(10):
        for k in range(points.shape[1]):
            points_[:,k] += 0.1
        pos_list.append(points_.copy())

    def init():
        return point_arts,

    def update(i):
        pos = pos_list[i]
        point_arts.set_data(pos[:,0], pos[:,1])
        point_arts.set_3d_properties(pos[:,2])
        return point_arts,

    ani = FuncAnimation(fig, update, frames=len(pos_list),
                        init_func=init, blit=True)

    plt.show()



"""
    Theory.SolidSphere.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.integrate import quad
from scipy.special import spherical_jn
from scipy.interpolate import UnivariateSpline
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt

def f_imp(s, R):

    def psi(r):
        return 0 if r > R else 1

    # note that np.sinc(x) is normalized. i.e., np.sin(np.pi*x)/(np.pi*x)
    return 4*np.pi*quad(lambda x: psi(x)*np.sinc(s*x/np.pi)*x**2, 0, R)[0]

def f(qv, R):
    scale = f_imp(0, R)
    y = np.zeros_like(qv)
    for k, q in enumerate(qv):
        y[k] = f_imp(q, R)
    return y/scale

def phi(s, R):
    sr = s*R
    return 3*np.power(sr, -3)*(np.sin(sr) - sr*np.cos(sr))

def phi_j1(s, R):
    sr = s*R
    # print(spherical_jn(1, 1e-6)/1e-6)
    # return 4*np.pi*R**2/s*spherical_jn(1, sr)
    # spherical_jn(1, x)/x → 1/3 as x → 0
    return 3*spherical_jn(1, sr)/sr

def a_func(q, R):
    return phi(q, R)**2

def b_to_a_func(q, R):
    return -8*phi(q, 2*R)

def find_boundary(y, ratio):
    w = np.where(np.abs(y) > ratio)[0]
    return w[-1] if len(w) > 0 else len(y)-1

QV = None
B2_LIMIT_RATIO = 0.99
def get_boundary_params(Rg, qv=None):
    if qv is None:
        global QV
        if QV is None:
            QV = np.linspace(0.005, 0.5, 100)
        qv = QV
    R = np.sqrt(5/3)*Rg
    rq = -phi(qv, 2*R)
    spline = UnivariateSpline(qv, rq, k=4, s=0)
    roots = spline.derivative().roots()
    b1, b2 = roots[0:2]
    k = -np.log(1/B2_LIMIT_RATIO - 1)/(b2 - b1)
    return b1, b2, k

def get_boundary_params_simple(Rg):
    R = np.sqrt(5/3)*Rg
    b1 = np.pi/R
    b2 = b1*1.5
    k = -np.log(1/B2_LIMIT_RATIO - 1)/(b2 - b1)
    return b1, b2, k

def demo1():
    qv = np.linspace(0.005, 0.5, 100)

    fig, axes = plt.subplots(nrows=2, ncols=3, figsize=(21, 11))

    ax1, ax2, ax3 = axes[0,:]
    ax4, ax5, ax6 = axes[1,:]

    y1 = f(qv, 30)
    y2 = phi(qv, 30)
    y3 = phi_j1(qv, 30)
    ax1.plot(qv, y1)
    ax2.plot(qv, y2)
    ax3.plot(qv, y3)

    ay1 = f(qv, 30)**2
    ay2 = phi(qv, 30)**2
    ay3 = phi_j1(qv, 30)**2

    for ax in [ax4, ax5, ax6]:
        ax.set_yscale('log')

    ax4.plot(qv, ay1)
    ax5.plot(qv, ay2)
    ax6.plot(qv, ay3)

    fig.tight_layout()
    plt.show()

def plot_boundary(ax, Rg, qv=None, boundary_ratio=0.05, fontsize=None):
    if qv is None:
        qv = np.linspace(0.005, 0.5, 100)
    RRg_ratio = np.sqrt(5/3)
    R = Rg*RRg_ratio
    ax.set_title("Rg=%g (R=%.3g)" % (Rg, R), fontsize=fontsize)
    ax.set_yscale('log')
    ax.plot(qv, a_func(qv, R), label='A(q)')
    axt = ax.twinx()
    axt.grid(False)
    axt.set_ylim(-2, 2)
    by = b_to_a_func(qv, R)
    axt.plot(qv, by, label='B(q)/A(q)', color='pink')
    k = find_boundary(by, boundary_ratio)
    ymin, ymax = axt.get_ylim()
    axt.set_ylim(ymin, ymax)
    b = qv[k]
    axt.plot([b, b], [ymin, ymax], color='red', label='boundary: q=%.2g' % b)

    ax.legend(fontsize=fontsize)
    axt.legend(bbox_to_anchor=(1, 0.8), loc='upper right', fontsize=fontsize)

def demo2():
    qv = np.linspace(0.005, 0.5, 100)
    nrows = 3
    ncols = 5
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(21, 11))

    fig.suptitle("Rank Boundary Estimation based on Solid Sphere Model", fontsize=20)

    for i in range(nrows):
        Rg_base = (i+1)*20
        for j in range(ncols):
            ax = axes[i,j]
            Rg = Rg_base + j*4
            plot_boundary(ax, Rg, qv)

    fig.tight_layout()
    fig.subplots_adjust(top=0.9)

    plt.show()

def demo3(Rg=32):
    qv = np.linspace(0.005, 0.5, 100)

    R = np.sqrt(5/3)*Rg
    aq = phi(qv, R)**2
    rq = -phi(qv, 2*R)
    bq = aq*rq

    b1, b2, k = get_boundary_params(Rg)

    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(14, 7))
    fig.suptitle("Adjusting the Logistic Function Parameter K to B(q)/A(q) Decline", fontsize=20)

    ax1.set_yscale('log')
    ax1.set_title("Rg=%.3g" % Rg, fontsize=16)
    ax1.plot(qv, aq)

    axt = ax1.twinx()
    axt.grid(False)

    for ax in [axt, ax2]:
        ax.plot(qv, rq, color='yellow', label='B(q)/A(q)')
        ax.plot(qv, bq, color='cyan', label='B(q)')
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax)
        ax.plot([b1, b1], [ymin, ymax], ':', color='red', label='q s.t. w(q)=0.5')
        ax.plot([b2, b2], [ymin, ymax], ':', color='blue', label='q s.t. 1 - w(q)=0.01')

    ax2.set_title("k=%.3g" % k, fontsize=16)
    ax2t = ax2.twinx()
    ax2t.grid(False)

    y = 1/(1 + np.exp(-k*(qv-b1)))
    ax2t.plot(qv, y, label='w(q)')
    ax2t.plot(qv, 1-y, label='1 - w(q)')
    ax2t.legend(bbox_to_anchor=(1, 0.7), loc='center right', fontsize=16)
    ax2.legend(bbox_to_anchor=(1, 0.4), loc='center right', fontsize=16)

    fig.tight_layout()
    fig.subplots_adjust(top=0.85)

    plt.show()

def demo4(Rg=35):
    R = np.sqrt(5/3)*Rg
    qv = np.linspace(0.005, 0.3, 100)
    b1, b2, _ = get_boundary_params_simple(Rg)

    fig, ax = plt.subplots(figsize=(10,7))
    ax.set_yscale('log')
    axt = ax.twinx()
    axt.grid(False)

    ax.set_title("Periodic appearances in $ \Phi(qR)^2 $ and $cos(qR)^2$", fontsize=20)

    ax.plot(qv, phi(qv, R)**2, label='$ \Phi(qR)^2 $')
    axt.plot(qv, np.cos(qv*R), ':', color='C1', label='$cos(qR)$')
    axt.plot(qv, np.cos(qv*R)**2, ':', color='C2', label='$cos(qR)^2$')

    ymin, _ = axt.get_ylim()
    ymax = 4
    axt.set_ylim(ymin, ymax)
    for b, label in [[b1, r"$b_1=\pi/R$"], [b2, r"$b_2=1.5\pi/R$"]]:
        axt.plot([b, b], [ymin, ymax], ':', color='red', label=label)

    ax.legend(fontsize=16)
    axt.legend(bbox_to_anchor=(1, 0.9), loc='upper right', fontsize=16)
    fig.tight_layout()
    plt.show()

class SolidSphere:
    def __init__(self):
        pass

    def fit(self, qv, Ic, initRg):
        scale = np.sqrt(5/3)
        initR = scale*initRg

        def obj_func(p):
            C, R = p
            return np.sum((Ic - C*phi(qv,R)**2)**2)

        res = minimize(obj_func, (1, initR))
        self.C, self.R = res.x
        self.Rg = self.R/scale
        return res.x

    def intensity(self, qv):
        return self.C * phi(qv,self.R)**2



# coding: utf-8
"""
    Shutdown.py

    from
        How to shutdown a computer using Python
        https://stackoverflow.com/questions/34039845/how-to-shutdown-a-computer-using-python

"""
import sys
import os

def shutdown_machine():
    if sys.platform == 'win32':
        os.system("shutdown /s /t 1")
    else:
        os.system('sudo shutdown now')

if __name__ == '__main__':
    shutdown_machine()



# coding: utf-8
"""

    Rank.RankView.py

    Copyright (c) 2022, SAXS Team, KEK-PF

"""
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from MenuButton import MenuButton
from DataUtils import get_in_folder
from .SrrTutor import rankview_plot

class RankView(Dialog):
    def __init__(self, parent, sd, data_type, **kwargs):
        self.sd = sd
        self.data_type = data_type
        self.kwargs = kwargs
        Dialog.__init__(self, parent, "Rank View", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        sd = self.sd
        data_type = self.data_type
        kwargs = self.kwargs
        xr_type = data_type.lower().find("x") >= 0
        if xr_type:
            M, E, vec, ecurve = sd.get_xr_data_separate_ly()
        else:
            M, _, vec, ecurve = sd.get_uv_data_separate_ly()
        self.M = M
        self.vec = vec

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(20,9))
        in_folder = get_in_folder()
        kwargs["return_axes"] = True
        self.axes_info = rankview_plot(fig, in_folder, sd, M, vec, data_type, xr_type, **kwargs)

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()

    def buttonbox(self):
        frame = Tk.Frame(self)
        frame.pack(fill=Tk.X)
        tframe = Tk.Frame(frame)
        tframe.pack(side=Tk.LEFT, padx=20)
        bframe = Tk.Frame(frame)
        bframe.pack(side=Tk.RIGHT, padx=20)

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        w = Tk.Button(bframe, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=100, pady=5)

        w = Tk.Button(bframe, text="Synchronize", width=10, command=self.synchronize)
        w.pack(side=Tk.LEFT, padx=100, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def synchronize(self):
        print("synchronize")
        ax0, axes = self.axes_info[0:2]
        kwargs = dict(azim=ax0.azim, elev=ax0.elev)
        for ax_row in axes:
            for ax in ax_row:
                ax.view_init(**kwargs)
        self.mpl_canvas.draw()

class RankViewMenu(Tk.Frame):
    def __init__(self, parent, dialog):
        Tk.Frame.__init__(self, parent)
        self.parent = parent
        self.dialog = dialog
        self.view_kw = None
        self.trimming = True
        self.menu = MenuButton(self, "Rank View", [
                            ("UV Rank View", lambda: self.show_rankview("UV")),
                            ("UV Rank View (corrected)", lambda: self.show_rankview("UV", correct=True)),
                            ("Xray Rank View", lambda: self.show_rankview("Xray")),
                            ("Xray Rank View (corrected)", lambda: self.show_rankview("Xray", correct=True)),
                            ("Change View Parameters", self.show_view_params_dialog),
                            ])
        self.menu.pack()
        self.menu.entryconfig(1, state=Tk.DISABLED)

    def config(self, **kwargs):
        self.menu.config(**kwargs)

    def show_rankview(self, data_type, correct=False):
        sd = self.dialog.serial_data
        pre_recog = self.dialog.pre_recog
        if correct:
            from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
            sd = get_corrected_sd_impl(sd, sd, pre_recog)
            data_label = "corrected "
        else:
            data_label = "no correction "
        dialog = RankView(self.dialog, sd, data_type, pre_recog=pre_recog, trim=self.trimming, data_label=data_label, view_kw=self.view_kw)
        dialog.show()

    def show_view_params_dialog(self):
        dialog = RankViewParamsDialog(self.dialog, view_kw=self.view_kw, trim=self.trimming)
        dialog.show()
        if dialog.applied:
            self.trimming, self.view_kw = dialog.get_parameters()

class RankViewParamsDialog(Dialog):
    def __init__(self, parent, view_kw=None, trim=True):
        self.view_kw = view_kw
        self.trim = trim
        self.applied = False
        Dialog.__init__(self, parent, "Rank View Parameters", visible=False, location='center right')

    def show(self):
        self._show()

    def body(self, body_frame):

        iframe = Tk.Frame(body_frame)
        iframe.pack(padx=20, pady=10)

        for j in range(3):
            iframe.columnconfigure(j, weight=1)

        grid_row = 0
        space = Tk.Label(iframe, text="    ")
        space.grid(row=grid_row, column=0)

        label = Tk.Label(iframe, text="■  Trimming Option")
        label.grid(row=grid_row, column=0, columnspan=3, pady=5, sticky=Tk.W)

        grid_row += 1
        self.trimming = Tk.IntVar()
        self.trimming.set(int(self.trim))
        cb = Tk.Checkbutton(iframe, text="apply automatic trimming", variable=self.trimming)
        cb.grid(row=grid_row, column=1, columnspan=2)

        grid_row += 1
        space = Tk.Label(iframe, text="    ")
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(iframe, text="■  3D View Parameters")
        label.grid(row=grid_row, column=0, columnspan=3, pady=5, sticky=Tk.W)

        azim, elev = (-60, 30) if self.view_kw is None else (self.view_kw["azim"], self.view_kw["elev"])

        grid_row += 1
        self.azim = Tk.IntVar()
        self.azim.set(azim)
        label = Tk.Label(iframe, text="azimuth: ")
        label.grid(row=grid_row, column=1, sticky=Tk.E)
        entry = Tk.Entry(iframe, textvariable=self.azim, width=6, justify=Tk.CENTER)
        entry.grid(row=grid_row, column=2)

        grid_row += 1
        self.elev = Tk.IntVar()
        self.elev.set(elev)
        label = Tk.Label(iframe, text="elevation: ")
        label.grid(row=grid_row, column=1, sticky=Tk.E)
        entry = Tk.Entry(iframe, textvariable=self.elev, width=6, justify=Tk.CENTER)
        entry.grid(row=grid_row, column=2)

    def apply(self):
        self.applied = True

    def get_parameters(self):
        return self.trimming.get()==1, dict(azim=self.azim.get(), elev=self.elev.get())



"""
    PhaseRetrieval.py

    Copyright (c) 2020-2023, SAXS Team, KEK-PF
"""
import copy
import numpy as np
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D

def plot_3d_image(ax, image, cmap=cm.plasma, min_value=None):
    if min_value is None:
        image_ = image
    else:
        image_ = copy.deepcopy(image)
        image_[ncp.logical_not(ncp.isfinite(image_))] = 0
        image_[image_ < min_value] = 0

    ax.voxels(image_, cmap=cmap)

def plot_3d_scatter(ax, image, cmap=cm.plasma, min_value=0, shape_limits=False):
        w = np.where(image > min_value)
        wi = np.array(w, dtype=int).T
        n = image.shape[0]
        xyz = wi - n/2
        print(xyz.shape)
        v = image[w]
        sc = ax.scatter3D(xyz[:,0], xyz[:,1], xyz[:,2], c=v, cmap=cmap, alpha=1, s=30)
        ax.get_figure().colorbar(sc, ax=ax)
        if shape_limits:
            h = n/2
            ax.set_xlim(-h, h)
            ax.set_ylim(-h, h)
            ax.set_zlim(-h, h)

"""
    originally got from
        https://github.com/tuelwer/phase-retrieval
        Phase retrieval: Fienup's algorithms

    ----------------------------------------------------------------------
    Implementation of Fienup's phase-retrieval methods. This function
    implements the input-output, the output-output and the hybrid method.
    
    Note: Mode 'output-output' and beta=1 results in 
    the Gerchberg-Saxton algorithm.
    
    Parameters:
        mag: Measured magnitudes of Fourier transform
        mask: Binary array indicating where the image should be
              if padding is known
        beta: Positive step size
        steps: Number of iterations
        mode: Which algorithm to use
              (can be 'input-output', 'output-output' or 'hybrid')
        verbose: If True, progress is shown
    
    Returns:
        x: Reconstructed image
    
    Author: Tobias Uelwer
    Date: 30.12.2018
    
    References:
    [1] E. Osherovich, Numerical methods for phase retrieval, 2012,
        https://arxiv.org/abs/1203.4756
    [2] J. R. Fienup, Phase retrieval algorithms: a comparison, 1982,
        https://www.osapublishing.org/ao/abstract.cfm?uri=ao-21-15-2758
    [3] https://github.com/cwg45/Image-Reconstruction
    ----------------------------------------------------------------------

    significantly mofified by SAXS Team, KEK-PF
    where the only skelton with mode = 'hybrid' is left.
"""
def fienup_phase_retrieval(mag, mask=None, beta=0.8, gpu=False, roll=True,
                            min_value=None, lim_value=0, saxs_curve=None,
                           steps=200, verbose=True):
    assert beta > 0, 'step size must be a positive number'
    assert steps > 0, 'steps must be a positive number'

    gpu_ok = False
    if gpu:
        try:
            import cupy as ncp
            gpu_ok = True
            if mask is not None:
                mask = ncp.asarray(mask)
        except:
            pass

    if gpu_ok:
        print('Using GPU')
        mag = ncp.asarray(mag)
    else:
        import numpy as ncp

    if mask is None:
        mask = ncp.ones(mag.shape)

    assert mag.shape == mask.shape, 'mask and mag must have same shape'

    # sample random phase and initialize image x 
    y_hat = mag*ncp.exp(1j*2*ncp.pi*ncp.random.rand(*mag.shape))
    # y_hat === rho in DENSS
    x = ncp.zeros(mag.shape)

    # previous iterate
    x_p = None
    x_hat_p = None

    # main loop
    for i in range(1, steps+1):
        # show progress
        if i % 100 == 0 and verbose: 
            print("step", i, "of", steps)

        # inverse fourier transform
        y = ncp.real(ncp.fft.ifftn(y_hat))

        # previous iterate
        if x_p is None:
            x_p = y
        else:
            x_p = x

        # updates for elements that satisfy object domain constraints
        x = y

        # find elements that violate object domain constraints 
        # or are not masked
        indices = ncp.logical_or(ncp.logical_and(y<lim_value, mask), 
                                ncp.logical_not(mask))

        # updates for elements that violate object domain constraints
        x[indices] = x_p[indices]-beta*y[indices]

        # fourier transform
        x_hat = ncp.fft.fftn(x)

        # satisfy fourier domain constraints
        # (replace magnitude with input magnitude)
        if saxs_curve is None:
            y_hat = mag*ncp.exp(1j*ncp.angle(x_hat))
        else:
            assert False

    if gpu_ok:
        x = ncp.asnumpy(x)

    if roll:
        try:
            from scipy import ndimage
            if min_value is None:
                x_ = x
            else:
                x_ = copy.deepcopy(x)
                x_[x_ < min_value] = 0
            com = ndimage.measurements.center_of_mass(x_)   # not supported in cupy
            m = np.array(com, dtype=int)
            print('com, m=', com, m)
            for k in range(x.ndim):
                m_ = x.shape[k]//2 - m[k]
                print([k], 'roll', m_)
                x = np.roll(x, m_, axis=k)
        except:
            print('roll failed.')

    return x



"""
    Models.Stochastic.DispersiveMonopore.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder
from molass_legacy.Models.Stochastic.ParamLimits import USE_K, MAX_PORESIZE
from molass_legacy.Models.Stochastic.RgReliability import determine_unreliables
from molass_legacy.Models.Stochastic.DispersivePdf import N0

M1_WEIGHT = 6
M2_WEIGHT = 4

assert USE_K

def guess_params_using_moments(x, y, egh_moments_list, peak_rgs, qualities, props, trust_max_num=None,
                               exec_spec=None, debug=False):
    from importlib import reload
    from molass_legacy.Models.Stochastic.ParamLimits import BASINHOPPING_SCALE
    import Models.Stochastic.DispersiveLimits
    reload(Models.Stochastic.DispersiveLimits)
    from molass_legacy.Models.Stochastic.DispersiveLimits import DispersiveParamsScaler
    import Models.Stochastic.DispersiveUtils
    reload(Models.Stochastic.DispersiveUtils)
    from molass_legacy.Models.Stochastic.DispersiveUtils import NUM_SDMCOL_PARAMS, NUM_SDMCUV_PARAMS, compute_elution_curves, investigate_sdm_params
    import Models.Stochastic.MobileDispersion
    reload(Models.Stochastic.MobileDispersion)
    from molass_legacy.Models.Stochastic.MobileDispersion import DispersionRatio

    PENALTY_SCALE = 1e3

    me = 1.5
    mp = 1.5

    num_components = len(peak_rgs)
    in_folder = get_in_folder()
    poresize_bounds = None
    unreliable_indeces = None

    if exec_spec is None:
        if in_folder.find('data01') > 0:
            unreliable_indeces = [1,2]
            poresize_bounds = 80, 120
            N0 = 10000
        elif in_folder.find('data02') > 0:
            unreliable_indeces = [1,2]
            poresize_bounds = 80, 120
            N0 = 5000
        elif in_folder.find('BSA_201') > 0:
            unreliable_indeces = [0]
            poresize_bounds = 75, 80
            # N0 = 50000      # best with tI-bound - 2000
            N0 = 50000
        elif in_folder.find('BSA_202') > 0:
            unreliable_indeces = [0]
            poresize_bounds = 75, 80
            # N0 = 14400/4
            N0 = 7000       # best with N0 = 50000, tI-bound - 2000
        elif in_folder.find('FER_OA_301') > 0:
            unreliable_indeces = [0,2]
            poresize_bounds = 75, 80
            # N0 = 50000      # best with tI-bound - 2000
            N0 = 50000
        elif in_folder.find('FER_OA_302') > 0:
            unreliable_indeces = [0,2]
            poresize_bounds = 75, 80
            # N0 = 14400/4
            N0 = 7000       # best with N0 = 50000, tI-bound - 2000
        elif in_folder.find('ALD_OA001') > 0:
            unreliable_indeces = [1]
            poresize_bounds = 75, 80
            N0 = 50000
        elif in_folder.find('ALD_OA002') > 0:
            unreliable_indeces = [1]
            poresize_bounds = 75, 80
            N0 = 50000
        else:
            unreliable_indeces = []
            poresize_bounds = 75, 80
            N0 = 50000
    else:
        unreliable_indeces = exec_spec['unreliable_indeces']
        poresize_bounds = exec_spec['poresize_bounds']
        N0 = exec_spec['init_N0']

    if unreliable_indeces is None:
        unreliable_indeces = determine_unreliables(peak_rgs, qualities, props, trust_max_num=trust_max_num, debug=True)

    print("unreliable_indeces=", unreliable_indeces)
    num_unreliables = len(unreliable_indeces)
    use_basinhopping = True         # it seems that global optimization is required for this problem
    params_scaler = DispersiveParamsScaler(egh_moments_list, 0, peak_rgs, unreliable_indeces,
                                           poresize_bounds=poresize_bounds)
    # 
    # bounds = [(300, 2000), (300, 5000), (x0-500, x0+500), (0, MAX_PORESIZE), (10000, 1e6), (-500, x0)]
    bounds = params_scaler.get_bounds()
    N = np.average(bounds[0])
    NT = np.average(bounds[1])
    # x0 = np.min([M[0] - np.sqrt(M[1])*10 for M in egh_moments_list])
    x0 = np.average(bounds[2])
    poresize = np.average(bounds[3])
    mdscale = np.sqrt(np.prod(bounds[4]))
    sigma0 = np.sqrt(np.percentile([M[1] for M in egh_moments_list], 30))
    t0 = min(500, np.sqrt(N0)*sigma0)
    print("t0=", t0)
    tI = x0 - t0
    init_colparams = np.concatenate([[N, NT, x0, poresize, tI], peak_rgs[unreliable_indeces]])
    target_moments_list = [(M[0], np.sqrt(M[1])) for M in egh_moments_list]
    zeros_rg_diff = np.zeros(num_components-1)

    def optimize_colparams(init_colparams):
        N, NT, x0, poresize, tI = init_colparams[0:NUM_SDMCOL_PARAMS]

        # dratio_spline = DispersionRatio(10, poresize*0.8, N, NT/N, poresize, t0, N0)

        print("init_colparams=", init_colparams)

        temp_rgs = np.asarray(peak_rgs).copy()
        def column_objective(p, debug=False, title=None):
            if use_basinhopping:
                p = params_scaler.scale_back(p)
            N_, K_, x0_, poresize_, tI_ = p[0:NUM_SDMCOL_PARAMS]
            p_rgs = p[NUM_SDMCOL_PARAMS:]
            temp_rgs[unreliable_indeces] = p_rgs
            penalty = PENALTY_SCALE * np.sum(np.min([zeros_rg_diff, temp_rgs[:-1] - temp_rgs[1:]], axis=0)**2)
            T_ = K_/N_
            rhov = temp_rgs/poresize_
            rhov[rhov > 1] = 1
            dev_list = []
            if debug:
                moments_list = []
            for k, (M, rho) in enumerate(zip(target_moments_list, rhov)):
                M1_ = x0_ + N_ * T_ * (1 - rho)**(me + mp)
                # t0_ = x0_ - tI_
                # t0_ + M1_ - x0_  + ... = m1_ - tI_ + ...
                # M2_ = np.sqrt(2 * N_ * T_**2 * (1 - rho)**(me + 2*mp) * (1 + (mdscale_*dratio_spline(temp_rgs[k]))**2))
                M2_ = np.sqrt(2 * N_ * T_**2 * (1 - rho)**(me + 2*mp) + (M1_ - tI_)**2/N0)
                dev_list.append(M1_WEIGHT*(M1_ - M[0])**2 + M2_WEIGHT*(M2_- M[1])**2)
                if debug:
                    print([k], "M1, M1_ = %.4g, %.4g" % (M[0], M1_))
                    print([k], "M2, M2_ = %.3g, %.3g" % (M[1], M2_))
                    moments_list.append([M1_, M2_])

            if debug:
                from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf
                print("moments_list=", moments_list)
                print("dev_list=", dev_list)
                print("p=", p)
                print("N_, T_, N0, tI_=", N_, T_, N0, tI_)
                with plt.Dp():
                    fig, ax = plt.subplots()
                    if title is not None:
                        ax.set_title(title)
                    ax.plot(x, y)
                    def plot_moments(ax, moments, color=None):
                        m1 = moments[0]
                        ax.axvline(x=m1, color=color)
                        s = moments[1]
                        ax.axvspan(m1-s, m1+s, color=color, alpha=0.3)

                    for moments, egh_moments in zip(moments_list, target_moments_list):
                        plot_moments(ax, moments, color='yellow')
                        plot_moments(ax, egh_moments, color='green')
                    ax.axvline(x=x0_, color='red')
                    scales = props * np.sum(y)
                    x_ = x - tI_
                    print("x_[0]=", x_[0])
                    t0_ = x0_ - tI_
                    for k, (rho, scale) in enumerate(zip(rhov, scales)):
                        ni_ = N_ * (1 - rho)**me
                        ti_ = T_ * (1 - rho)**mp
                        print([k], ni_, ti_, N0, t0_)
                        cy = scale*dispersive_monopore_pdf(x_, ni_, ti_, N0, t0_)
                        ax.plot(x, cy, ":")
 
                    rv = np.linspace(20, poresize, 100)
                    rhov_ = rv/poresize_
                    rhov_[rhov_ > 1] = 1
                    trv = x0_ + N_ * T_ * (1 - rhov_)**(me + mp)

                    axt = ax.twinx()
                    axt.grid(False)
                    axt.plot(trv, rv, color='yellow')
                    ehg_m1 = [M[0] for M in egh_moments_list]
                    axt.plot(ehg_m1, temp_rgs, 'o')

                    fig.tight_layout()
                    ret = plt.show()
                    if not ret:
                        if title is None:
                            assert False
                        return

            return np.sum(np.asarray(dev_list)*props) + penalty

        if use_basinhopping:
            init_colparams = params_scaler.scale(init_colparams)
        if debug:
            print("init_colparams=", init_colparams)
            ret = column_objective(init_colparams, debug=True, title="before column minimize")
            if ret is None:
                return

        if use_basinhopping:
            bounds_ = [(0, BASINHOPPING_SCALE)] * len(init_colparams)
            minimizer_kwargs = dict(method='Nelder-Mead', bounds=bounds_)
            res = basinhopping(column_objective, init_colparams, minimizer_kwargs=minimizer_kwargs)
        else:
            res = minimize(column_objective, init_colparams, bounds=bounds, method='Nelder-Mead')

        if debug:
            print("res.x=", res.x)
            ret = column_objective(res.x, debug=True, title="after column minimize")
            if ret is None:
                return
        if use_basinhopping:    
            res.x = params_scaler.scale_back(res.x)
        temp_rgs[unreliable_indeces] = res.x[NUM_SDMCOL_PARAMS:]

        init_scales = props * np.sum(y)
        # init_cuvparams = dratio_spline.covert_to_curveparams(peak_rgs, props, target_moments_list, res.x, init_scales, debug_info=(x, y))
        init_cuvparams = np.concatenate([res.x[:NUM_SDMCOL_PARAMS-1], [N0, res.x[NUM_SDMCOL_PARAMS-1]], init_scales])
        if init_cuvparams is None:
            return
        print("init_cuvparams=", init_cuvparams)

        if debug:
            ret = investigate_sdm_params(x, y, init_cuvparams, num_unreliables, temp_rgs, bounds)
            if not ret:
                return

        return res.x, temp_rgs, bounds, init_cuvparams

    # first trial
    temp_colparams = init_colparams
    for k in range(1):
        print([k], "temp_colparams=", temp_colparams)
        opt_info = optimize_colparams(temp_colparams)
        if opt_info is None:
            return
        temp_colparams, temp_rgs, bounds, init_cuvparams = opt_info

    init_scales = props * np.sum(y)
    cuvparams = init_cuvparams[0:NUM_SDMCUV_PARAMS]

    def elution_objective(p, debug=False, title=None):
        # p_rgs = p[NUM_SDMCOL_PARAMS:NUM_SDMCOL_PARAMS+num_unreliables]
        # temp_rgs[unreliable_indeces] = p_rgs
        cy_list, ty = compute_elution_curves(x, np.concatenate([cuvparams, p]), temp_rgs)

        if debug:
            x0_ = cuvparams[2]
            with plt.Dp():
                fig, ax = plt.subplots()
                if title is not None:
                    ax.set_title(title)
                ax.plot(x, y)
                for cy in cy_list:
                    ax.plot(x, cy, ":")
                ax.plot(x, ty, ":", color="red")
                ax.axvline(x=x0_, color='red')
                # ax.axvline(x=tI_, color='gray')
                fig.tight_layout()
                ret = plt.show()
                if not ret:
                    return

        return np.sum((y - ty)**2)

    if debug:
        print("init_scales=", init_scales)
        ret = elution_objective(init_scales, debug=True, title="before elution minimize")
        if ret is None:
            return

    scale_bounds = [(0, 10)]*len(init_scales)
    res = minimize(elution_objective, init_scales, bounds=scale_bounds, method='Nelder-Mead')
    
    if debug:
        print("res.x=", res.x)
        ret = elution_objective(res.x, debug=True, title="after elution minimize")
        if ret is None:
            return
    
    ret_params = np.concatenate([cuvparams, res.x])
    return ret_params, temp_rgs, bounds


"""
    SingleInstance.py

    borrowed from
        http://code.activestate.com/recipes/578453-python-single-instance-cross-platform/

    modified as follows:
        1) allow LOCK_PATH to be given by user
"""
import sys
import os

try:
    import fcntl
except ImportError:
    fcntl = None

OS_WIN = False
if 'win32' in sys.platform.lower():
    OS_WIN = True


class SingleInstance:
    def __init__(self, lock_path=None, raise_=True):
        if lock_path is None:
            lock_path = os.path.join(os.path.abspath(os.path.dirname(sys.argv[0])), "lock")
        self.lock_path = lock_path
        self.fh = None
        self.is_running = False
        try:
            self.do_magic()
        except:
            if raise_:
                raise
            else:
                self.is_running = None

    def do_magic(self):
        if OS_WIN:
            try:
                if os.path.exists(self.lock_path):
                    os.unlink(self.lock_path)
                self.fh = os.open(self.lock_path, os.O_CREAT | os.O_EXCL | os.O_RDWR)
            except EnvironmentError as err:
                if err.errno == 13:
                    self.is_running = True
                else:
                    raise

        else:
            try:
                self.fh = open(self.lock_path, 'w')
                fcntl.lockf(self.fh, fcntl.LOCK_EX | fcntl.LOCK_NB)
            except EnvironmentError as err:
                if self.fh is not None:
                    self.is_running = True
                else:
                    raise

    def clean_up(self):
        # this is not really needed
        try:
            if self.fh is not None:
                if OS_WIN:
                    os.close(self.fh)
                    os.unlink(self.lock_path)
                else:
                    fcntl.lockf(self.fh, fcntl.LOCK_UN)
                    self.fh.close() # ???
                    os.unlink(self.lock_path)
        except Exception as err:
            # logger.exception(err)
            raise # for debugging porpuses, do not raise it on production


if __name__ == "__main__":
    import time

    si = SingleInstance()
    try:
        if si.is_running:
            sys.exit("This app is already running!")
        time.sleep(20) # remove
        # do other stuff
    finally:
        si.clean_up()



# coding: utf-8
"""
    QrgLimitsAdjuster.py

    Copyright (c) 2016, Masatsuyo Takahashi, KEK-PF
"""

import sys
import traceback
import time
import numpy                as np
from Quality                import fit_consistency_with_stdev

DEBUG = False

class Struct:
    def __init__( self, **entries ): 
        self.__dict__.update(entries)

class ResultEvaluator:
    def __init__( self, result ):
        self.result = result
        self.has_max_qRg = False

    def is_valid( self, qrg_limits ):
        if self.result is None: return False
        if self.result.Rg is None: return False
        self.has_max_qRg = True
        if self.result.max_qRg > qrg_limits[1] :  return False
        return True

    def can_be_improved( self, qrg_limits ):
        if not self.has_max_qRg: return True
        if self.result.max_qRg < qrg_limits[1] - 0.1:
            return True
        else:
            return False

class QrgLimitsAdjuster:
    def __init__( self, fit, guinier, qrg_limits, result, min_num_points=10, max_num_points=40 ):
        self.fit                = fit
        self.f0                 = fit.f0
        self.guinier            = guinier
        self.qrg_limits         = qrg_limits
        self.result             = result
        self.modified           = False
        self.previous_interval  = None
        self.previous_wls       = None
        self.min_num_points     = min_num_points
        self.max_num_points     = max_num_points

        # print( 'QrgLimitsAdjuster: qrg_limits=', qrg_limits )

        if result is None:
            if DEBUG: print( 'QrgLimitsAdjuster: None return' )
            return

        if DEBUG:
            Rg = 0 if result.Rg is None else result.Rg
            print( 'QrgLimitsAdjuster: Rg=%.4g, fit.Rg=%.4g, Rg/fit.Rg=%.4g' % ( Rg, fit.Rg, Rg /fit.Rg ) )

        if ( result.Rg is not None
            and ( result.Aggregated is not None and abs( result.Aggregated ) > 0.5 )
            and result.max_qRg <= qrg_limits[1]
            ):
            # it is better not to try to extend max_qRg when anomalies exist.
            if DEBUG: print( 'QrgLimitsAdjuster: Anomaly return' )
            return

        if result.Rg is None or result.Rg > fit.Rg or result.Rg / fit.Rg > 0.98:
            # Rg / fit.Rg == 0.989 for 20160628/AIMdim01_00231_sub.dat
            # print( 'max_qRg=', result.max_qRg )
            self.try_loop( 'L', self.try_shifting_left )

        else:
            if DEBUG: print('TBD!!!: not ( result.Rg > fit.Rg or result.Rg / fit.Rg > 0.98 ) ')
            self.try_loop( 'R', self.try_extending_right, previous_ok=True )

        """
        try:
            self.minimize_lower_limit()
        except Exception as exc:
            print( exc )
        """

    def try_loop( self, direction, method, previous_ok=False ):
        result = self.result

        if DEBUG:
            print( 'try_loop: result=', result )
            if result is not None:
                print( 'try_loop: result.Rg=', result.Rg, ', result.max_qRg=', result.max_qRg )

        num_trials = self.max_num_points if result.Rg is None else 100

        n = 0

        last_better_result      = None
        last_better_consitency  = None
        evaluator = ResultEvaluator( result )
        if evaluator.is_valid( self.qrg_limits ):
            last_better_result      = result
            consistency             = fit_consistency_with_stdev( self.fit.Rg, result.Rg, result.Rg_stdev )
            last_better_consitency  = consistency

        while ( result is not None
                and ( result.Rg is None and n < num_trials
                    or result.max_qRg is not None
                        and (  result.max_qRg > self.qrg_limits[1]
                            or result.max_qRg < self.qrg_limits[1] - 0.1
                            or (   last_better_consitency is not None
                               and last_better_consitency < 0.7
                               )
                        )
                    )
                ):
            if DEBUG: print( 'try_loop: max_qRg=', result.max_qRg )
            if result.max_qRg is not None:
                stdev_ratio = result.Rg_stdev / result.Rg
                if DEBUG: print( 'stdev_ratio=', stdev_ratio )
                if stdev_ratio < 0.3:
                    if direction == 'L':
                        if result.max_qRg < self.qrg_limits[1] - 0.1:
                            break
                    else:
                        if result.max_qRg > self.qrg_limits[1] + 0.1:
                            break

            try:
                result = method( result )
                evaluator = ResultEvaluator( result )
                if evaluator.is_valid( self.qrg_limits ):
                    consistency = fit_consistency_with_stdev( self.fit.Rg, result.Rg, result.Rg_stdev )
                    if last_better_consitency is None or consistency > last_better_consitency:
                        last_better_result      = result
                        last_better_consitency  = consistency
                        if DEBUG: print( '---- fit.Rg=%.4g, Rg=%.4g, consistency=%.4g' % ( self.fit.Rg, result.Rg, consistency ) )
            except Exception as e:
                if DEBUG: print( 'Exception:', e.args )
                if len( e.args[0] ) != 2:
                    raise Exception( 'len( e.args[0] ) > 2' )
                    break
                    # TODO: do it better.

                f_, t_ = e.args[0][1:]
                result = Struct( Rg=None, max_qRg=None, From=f_, To=t_, result_type=-1, Quality=0 )
                if DEBUG: print( 'result.Rg=', result.Rg, ', result.From=', result.From, ', result.To=', result.To )
            n += 1

        result = last_better_result

        if self.modified and not hasattr( result, 'Quality' ):
            result = self.guinier.get_result( fit=self.fit, result=result )

        if result is not None:
            if DEBUG:
                if result.Rg is not None:
                    print(  'result_type=%s, Rg=%.4g, max_qRg=%.4g, Quality=%.4g'
                                % ( result.result_type, result.Rg, result.max_qRg, result.Quality )
                         )

            if result.result_type is not None and result.result_type >= 0:
                self.result = result

    def try_shifting_left( self, result ):
        f_, t_ = result.From, result.To
        size = t_ - f_ + 1
        if DEBUG: print( 'try_shifting_left: [%d, %d], size=%d' % ( f_, t_, size ) )
        if size < 10:
            f_ = t_ - 9
            if f_ < self.f0:
                return None
            size = t_ - f_ + 1
            assert( size >= self.min_num_points )
        else:
            if f_ > self.f0:
                f_ -= 1
                if size >= self.max_num_points:
                    # has enough points, so don't widen any more
                    t_ -= 1
            else:
                if size > self.min_num_points:
                    t_ -= 1
                else:
                    return None

        self.guinier.estimate_rg( [ f_, t_ ] )
        if self.guinier.wls.Rg is None:
            # restore the previous state
            self.guinier.estimate_rg( [ result.From, result.To ] )
            raise Exception( [ 'failed to calculate Rg', f_, t_ ] )

        Rg = self.guinier.wls.Rg
        Rg_stdev = self.guinier.wls.sigmaRg
        max_qRg = np.sqrt( self.guinier.x[t_] ) * Rg
        if DEBUG: print( 'try_shifting_left: [%d, %d], max_qRg=%.4g' % ( f_, t_, max_qRg ) )
        self.modified   = True
        result_type_ = self.result.result_type * 10 + 8
        return Struct( Rg=Rg, Rg_stdev=Rg_stdev, From=f_, To=t_, wls=self.guinier.wls, max_qRg=max_qRg, result_type=result_type_ )

    def try_extending_right( self, result ):
        f_, t_ = result.From, result.To
        size = t_ - f_ + 1
        if DEBUG: print( 'try_extending_right: [%d, %d], size=%d' % ( f_, t_, size ) )
        if size < 10:
            t_ = f_ + 9
            if t_ > self.fit.t1:
                return None
            size = t_ - f_ + 1
            assert( size >= self.min_num_points )
        else:
            if t_ < self.fit.t1 + self.min_num_points:
                if size >= self.max_num_points:
                    # has enough points, so don't widen any more
                    f_ += 1
                t_ += 1
            else:
                if f_ < self.fit.t1 and size > self.min_num_points:
                    f_ += 1
                else:
                    return None

        self.guinier.estimate_rg( [ f_, t_ ] )
        if self.guinier.wls.Rg is None:
            # restore the previous state
            self.guinier.estimate_rg( [ result.From, result.To ] )
            raise Exception( [ 'failed to calculate Rg', f_, t_ ] )

        Rg = self.guinier.wls.Rg
        Rg_stdev = self.guinier.wls.sigmaRg
        max_qRg = np.sqrt( self.guinier.x[t_] ) * Rg
        if DEBUG: print( 'try_extending_right: [%d, %d], max_qRg=%.4g' % ( f_, t_, max_qRg ) )
        self.modified   = True
        result_type_ = self.result.result_type * 10 + 9
        return Struct( Rg=Rg, Rg_stdev=Rg_stdev, From=f_, To=t_, wls=self.guinier.wls, max_qRg=max_qRg, result_type=result_type_ )

    def minimize_lower_limit( self ):
        print( 'minimize_lower_limit' )
        Rg  = self.result.Rg
        f   = self.result.From
        t   = self.result.To
        min_i = None
        start = time.time()
        for i in range( f ):
            try:
                self.guinier.estimate_rg( [ i, t ] )
                Rg_ = self.guinier.wls.Rg
                if Rg_ is None:
                    continue
                if abs( Rg_/Rg - 1 ) < 0.01:
                    min_i = i
                    break
            except:
                continue
        print( 'f0, f, min_i=', self.f0, f, min_i, time.time() - start )

    def get_adjusted_result( self ):
        if self.result is None:
            return None
        else:
            return self.result
            """
            if self.modified:
                result_type = self.result.result_type * 10 + 7
                return self.guinier.get_result( self.fit, result_type=result_type )
            else:
                # TODO: why is this different from the above?
                return self.result
            """



# coding: utf-8
"""
    EvolvingFactorAnalysis.py

    Copyright (c) 2018, Masatsuyo Takahashi, KEK-PF
"""
import copy
import numpy as np
import matplotlib.pyplot    as plt
import matplotlib.gridspec  as gridspec
from mpl_toolkits.mplot3d   import Axes3D
from matplotlib.collections import PolyCollection
from matplotlib             import colors as mcolors
from molass_legacy.Models.ElutionCurveModels     import EMGA, EGHA
from XrayDecomposer         import XrayDecomposer, proof_plot
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.BasicUtils import clear_dirs_with_retry
from molass_legacy.KekLib.NumpyUtils import np_savetxt

CONCENTRATION_NOMALIZE  = True
NUM_SVS_TO_SHOW         = 5
DRAW_IN_THE_ORDER_OF_SV = True

def rectify( n, pair ):
    ret = []
    for rec in pair:
        r = rec[0:n]
        ret.append( list(r) + [0]*(n-(len(r))) )
    return ret

class EvolvingFactorAnalysis:
    def __init__( self, data, ecurve, eslice, y, s_lim=None ):

        self.data   = data
        self.ecurve = ecurve
        self.eslice = eslice
        self.y      = y
        self.y_     = y_ = y if eslice is None else y[eslice]

        if False:
            dplt.plot( y )
            dplt.plot( y_ )
            dplt.show()

        if CONCENTRATION_NOMALIZE:
            ratio = 1/np.max( [ y_, np.ones( len(y_) )* 0.001 ], axis=0 )
            normalized_data = np.prod( [ data, ratio ], axis=0 )
        else:
            normalized_data = data

        # dslice = slice(2,6)
        # dslice = slice(0,2)
        dslice = slice(0,5)
        # dslice = slice(1,2)
        U, s, VT = np.linalg.svd( normalized_data )
        print( 's=', s[0:10] )
        denoised_data =np.dot(  np.dot( U[:,dslice], np.diag( s[dslice] ) ), VT[dslice, :] )
        U, s, VT = np.linalg.svd( denoised_data )
        print( 'denoised s=', s[0:10] )

        s_max = s[0]
        if s_lim is None:
            s_lim = s_max*0.2

        if not CONCENTRATION_NOMALIZE:
            s_lim_ratio = s_lim / s_max

        rank_list = []
        sv_list = []
        sv_list2 = []
        for t in range( 1, denoised_data.shape[1] ):
            M   = denoised_data[:,0:t]
            U, s, V = np.linalg.svd( M )

            s_ = s/s[0]

            sv_list.append( [ s, s_ ] )
            sv_list2.append( rectify(NUM_SVS_TO_SHOW, [ s, s_ ]) )
            if t < 20:
                print( [t], 's=', s[0:10] )

            if CONCENTRATION_NOMALIZE:
                s_lim_  = s_lim
            else:
                s_lim_ = s[0] * s_lim_ratio

            rank_list.append( len( np.where( s > s_lim_ )[0] ) )

        self.rank_list = rank_list
        self.sv_list = sv_list
        self.sv_array = np.array(sv_list2)
        print( 'self.sv_array.shape=', self.sv_array.shape )

    def show_plot( self, sd, in_folder, rank_boundaries=None ):

        model = EGHA()
        x   = self.ecurve.x
        y   = self.y
        fa  = XrayDecomposer( self.ecurve, x, y, sd.intensity_array, retry_valley=True, model=model, deeply=True )

        x_  = np.arange(self.eslice.start, self.eslice.stop)

        fig = plt.figure( figsize=( 24, 8 ) )
        gs = gridspec.GridSpec( 1, 3 )
        ax1 = fig.add_subplot( gs[0, 0] )
        ax2 = ax1.twinx()
        ax3 = fig.add_subplot( gs[0, 1], projection='3d' )
        ax4 = fig.add_subplot( gs[0, 2], projection='3d' )

        fontsize = 14

        plt.yticks(fontsize=fontsize)

        ax1.set_title( 'Evolving Factor Analysis for ' + in_folder, fontsize=fontsize )
        ax1.set_ylabel( 'Intensity', fontsize=fontsize )
        ax1.set_xlabel( 'Elution No.', fontsize=fontsize )
        ax2.set_ylabel( 'Matrix Rank', fontsize=fontsize )
        ax1.plot( x_, self.y_, color='orange', linewidth=5 )

        def draw_decomposition( ax, fit_recs ):

            x_residual = copy.deepcopy( self.y_ )

            for _, func, _ in fit_recs:
                ey = func( x_ )
                ax.plot( x_, ey )
                x_residual -= ey

            ax.plot( x_, x_residual )

        draw_decomposition( ax1, fa.fit_recs )

        ax2.plot( x_[1:], self.rank_list, linewidth=5, color='gray' )
        ax2.set_yticks([0, 1, 2])

        ax3.set_title( "Concentration-normalized Singular Values", fontsize=fontsize )
        ax3.set_xlabel( '\nElution No.', fontsize=fontsize )
        ax3.set_ylabel( '\nRank', fontsize=fontsize )
        ax3.set_zlabel( '\nScale', fontsize=fontsize )
        ax4.set_title( "Maxvalue-normalized Singular Values", fontsize=fontsize )
        ax4.set_xlabel( '\nElution No.', fontsize=fontsize )
        ax4.set_ylabel( '\nRank', fontsize=fontsize )
        ax4.set_zlabel( '\nScale Ratio', fontsize=fontsize )

        for ax in [ ax1, ax2, ax3, ax4 ]:
            ax.tick_params( labelsize=fontsize )

        start = self.eslice.start

        if DRAW_IN_THE_ORDER_OF_SV:

            limit = min( self.sv_array.shape[2], NUM_SVS_TO_SHOW )
            for i in range(limit):
                x   = np.arange( start+i, self.eslice.stop-1 )
                y   = np.ones(len(x)) * (i+1)
                x_  = x - start
                z   = self.sv_array[x_,0,i]
                ax3.plot( x, y, z, 'o', markersize=3, label='s%d' % i  )
                z_  = self.sv_array[x_,1,i]
                ax4.plot( x, y, z_, 'o', markersize=3, label='s%d' % i  )

            ax3.legend(fontsize=fontsize, loc='upper left')
            ax4.legend(fontsize=fontsize, loc='upper left')

        else:

            for k, sv in enumerate( self.sv_list ):
                limit = min( len(sv[0]), NUM_SVS_TO_SHOW ) 
                x   = np.ones(limit) * (start + k)
                y   = np.arange( limit ) + 1
                z   = sv[0][0:limit]
                ax3.plot( x, y, z, 'o', markersize=3 )
                z_  = sv[1][0:limit]
                ax4.plot( x, y, z_, 'o', markersize=3 )

        if rank_boundaries is not None:
            ymin, ymax = ax1.get_ylim()
            ax1.set_ylim( ymin, ymax )
            for b in rank_boundaries:
                ax1.plot( [b, b], [ymin, ymax ], color='red', linewidth=5 )

            for k, ax in enumerate([ ax3, ax4 ]):
                ymin, ymax = ax.get_ylim()
                zmin, zmax = ax.set_zlim()
                zs_list = rank_boundaries
                verts1  = []
                for b in rank_boundaries:
                    yp  = [ ymin, ymin, ymax, ymax ]
                    zp  = [ zmax, zmin, zmin, zmax ]
                    verts1.append( list( zip( yp, zp ) ) )
                    if DRAW_IN_THE_ORDER_OF_SV:
                        x   = np.ones(NUM_SVS_TO_SHOW)*b
                        y   = np.arange( NUM_SVS_TO_SHOW ) + 1
                        z   = self.sv_array[b-start,k,0:NUM_SVS_TO_SHOW]
                        ax.plot( x, y, z, ':', color='gray' )

                r = mcolors.to_rgba('red', alpha=0.2)
                poly1 = PolyCollection(verts1, facecolors=[r]*len(verts1) )
                # poly1.set_alpha(0.2)
                ax.add_collection3d(poly1, zs=zs_list, zdir='x')

        fig.tight_layout()
        plt.show()

class ProofData:
    def __init__( self, sd, mapper ):
        print( mapper.x_curve.peak_info )
        x_curve = mapper.x_curve
        a_baseline = mapper.a_base + mapper.a_base_adjustment
        x_baseline = mapper.x_base + mapper.x_base_adjustment

        x_corrected = mapper.x_vector - x_baseline
        sd.apply_baseline_correction( mapper.get_mapped_info() )
        data    = sd.intensity_array

        print( 'sd.xray_slice=', sd.xray_slice )
        self.xray_slice = sd.xray_slice

        y   = x_corrected
        x   = np.arange( len(y) )
        emg_model = EMGA()
        self.fa = XrayDecomposer( x_curve, x, y, data, retry_valley=True, model=emg_model, deeply=True )
        self.ridges = [ data[int(info[1]+0.5),:,1] for info in x_curve.peak_info ]
        assert len( self.ridges ) == len( self.fa.fit_recs )

        self.qvector = sd.qvector
        self.evector = np.zeros( len(self.qvector) )
        self.C  = np.array( [ rec[1](x) for rec in self.fa.fit_recs ] )

        if False:
            for ridge in self.ridges:
                dplt.plot( ridge )
            dplt.show()

    def show( self, parent ):
        proof_plot( self.fa, parent )

    def save_weak( self, folder ):
        clear_dirs_with_retry( [folder] )

        A   = np.array( self.ridges ).T
        M   = np.dot( A, self.C )
        for j in range( M.shape[1] ):
            file = folder + '/Weak_%05d.dat' % j
            data = np.array( [ self.qvector, M[:,j], self.evector ] ).T
            np_savetxt( file, data )

    def save_strong( self, folder ):
        clear_dirs_with_retry( [folder] )

        shift_amount = 20
        zero_slice_ = slice( self.xray_slice.start + shift_amount, self.xray_slice.stop + shift_amount )

        strong_ridges = []
        for k, ridge in enumerate(self.ridges):
            if k % 2 == 0:
                ridge_ = ridge
            else:
                ridge_ = ridge
                ridge_[zero_slice_] = 0
            strong_ridges.append( ridge_ )

        A   = np.array( strong_ridges ).T
        M   = np.dot( A, self.C )
        for j in range( M.shape[1] ):
            file = folder + '/Strong_%05d.dat' % j
            data = np.array( [ self.qvector, M[:,j], self.evector ] ).T
            np_savetxt( file, data )



# coding: utf-8
"""

    ExistingNmf.py

    Copyright (c) 2019, SAXS Team, KEK-PF

"""
import copy
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
from ThreeDimUtils import compute_plane
from LmfitThreadSafe import minimize, Parameters
from ModeledData import ModeledData, simple_plot_3d
import molass_legacy.KekLib.DebugPlot as plt

class ExistingNmf:
    def __init__(self, M, module='nimfa'):
        if True:
            import nimfa
            tool = "Nimfa"
            if False:
                lsnmf = nimfa.Lsnmf(M, seed='random_vcol', rank=2, max_iter=100)
                lsnmf_fit = lsnmf()
                W = np.asarray(lsnmf_fit.basis())
                H = np.asarray(lsnmf_fit.coef())
            else:
                pmf = nimfa.Pmf(M, seed='random_vcol', rank=2, max_iter=100)
                pmf_fit = pmf()
                W = np.asarray(pmf_fit.basis())
                H = np.asarray(pmf_fit.coef())
        else:
            from sklearn.decomposition import NMF
            tool = "Scikit-Learn"
            model = NMF(n_components=2, init='random', random_state=0)
            W = model.fit_transform(M)
            H = model.components_

        self.tool = tool
        self.H = H
        self.W = W

class ExistingNmfDemo:
    def __init__(self, data, module='nimfa'):
        M = copy.deepcopy(data)
        M[M<0] = 0

        n_angles = M.shape[0]
        n_elutions = M.shape[1]

        if True:
            nmf = ExistingNmf(M, module=module)
            W = nmf.W
            H = nmf.H
            tool = nmf.tool
        else:
            import pymf
            nmf = pymf.NMF(M, num_bases=2)
            nmf.factorize()
            W = nmf.W
            H = nmf.H
            tool = 'PyMF'

        if True:
            from DataUtils import get_in_folder

            print('W=', W.shape)
            print('H=', H.shape)

            fig = plt.figure(figsize=(14,7))
            in_folder = get_in_folder()
            fig.suptitle("Non-negative Matrix Factorization using " + tool + " for " + in_folder, fontsize=20)

            ax1 = fig.add_subplot(121, projection='3d')
            ax2 = fig.add_subplot(122)

            simple_plot_3d(ax1, M, color='cyan', alpha=0.2)

            x = np.ones(n_elutions)*0.02
            y = np.arange(n_elutions)
            for k in range(H.shape[0]):
                z = H[k,:]
                ax1.plot(x, y, z)
                ax2.plot(y, z)

            fig.tight_layout()
            fig.subplots_adjust(top=0.9)
            plt.show()

def sklearn_nmf_spike():
    qsize = 600
    qvector = np.linspace(0.01, 0.6, qsize)
    n_elutions = 300
    pd = ModeledData(qvector, n_elutions)
    M = pd.get_data(debug=False)

    model = NMF(n_components=1, init='random', random_state=0)
    W = model.fit_transform(M)
    H = model.components_
    print('W=', W.shape)
    print('H=', H.shape)
    print('W@H=', (W@H).shape)

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    simple_plot_3d(ax, M)
    x = np.ones(n_elutions)*0.02
    y = np.arange(n_elutions)
    z = H[0,:]
    ax.plot(x, y, z, color='orange')
    x = np.arange(qsize)
    y = np.ones(qsize)*150
    z = W[:,0]
    ax.plot(x, y, z, color='green')
    fig.tight_layout()
    plt.show()


def nimfa_spike():
    import nimfa

    qsize = 600
    qvector = np.linspace(0.01, 0.6, qsize)
    n_elutions = 300
    pd = ModeledData(qvector, n_elutions)
    M = pd.get_data(debug=False)

    lsnmf = nimfa.Lsnmf(M, seed='random_vcol', rank=1, max_iter=100)
    lsnmf_fit = lsnmf()

    W = np.asarray(lsnmf_fit.basis())
    H = np.asarray(lsnmf_fit.coef())

    print('W=', W.shape)
    print('H=', H.shape, type(H))

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    simple_plot_3d(ax, M)
    x = np.ones(n_elutions)*0.02
    y = np.arange(n_elutions)
    z = H[0,:]
    scale = np.max(z)
    print('z=', z.shape)
    ax.plot(x, y, z/scale, color='orange')

    x = np.arange(qsize)
    y = np.ones(qsize)*150
    z = W[:,0]
    print('z=', z.shape)
    ax.plot(x, y, z*scale, color='green')
    fig.tight_layout()
    plt.show()

class MatrixFactorization:
    def __init__(self, sd):
        pass



"""
    ElutionCurveModels.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF
"""
from molass_legacy.Models.ModelUtils import model_debug_plot
from molass_legacy.Models.EGH import EGH, EGHA, egh, egha, egh_x_from_height_ratio
from molass_legacy.Models.EMG import EMG, EMGA, emg, emga


"""
    Optimizer.XrUvScaleRatio.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np

ALLOW_RATIO = 1.5       # > 1.0 for OA_Ald
PENALTY_SCALE = 0.1     # too large value may cause the optimization early stages to fail

def xruv_scale_ratio_penalty(xr_scales, uv_scales, debug=False):
    uv_xr_ratio = uv_scales/xr_scales
    average = np.average(uv_xr_ratio)
    std = np.std(uv_xr_ratio)
    dev = np.abs(uv_xr_ratio - average)
    exv = dev[dev > std*ALLOW_RATIO]
    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        print("average=", average)
        print("dev=", dev)
        allow = std*ALLOW_RATIO
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("UV/Xray Ratio Debug", fontsize=16)
            ax.bar(np.arange(len(uv_xr_ratio)), uv_xr_ratio)
            ax.axhline(average)
            ax.axhspan(average - allow, average + allow, color="yellow", alpha=0.3)
            fig.tight_layout()
            plt.show()
    if len(exv) > 0:
        penalty = PENALTY_SCALE*np.sum(exv)
    else:
        penalty = 0.0
    return penalty


"""
    ErrorResult.py

    Copyright (c) 2016-2023, SAXS Team, KEK-PF
"""
class ErrorFitResult:
    def __init__( self ):
        self.aic    = None
        self.bic    = None

class ErrorFit:
    def __init__( self ):
        self.Rg     = None
        self.degree = None
        self.result = ErrorFitResult()
        self.I0     = None

class ErrorResult:
    def __init__( self ):
        self.Rg         = None
        self.Rg_        = 0
        self.Rg_stdev   = 0
        self.I0         = None
        self.I0_        = 0
        self.I0_stdev   = 0
        self.From       = None
        self.To         = None
        self.min_q      = None
        self.max_q      = None
        self.min_qRg    = None
        self.max_qRg    = None
        self.fit        = ErrorFit()
        self.bico_mono_ratio    = None
        self.IpI            = None
        self.bicomponent    = None
        self.head_trend     = None
        self.result_type    = None
        self.quality_object = None
        self.basic_quality  = 0
        self.Quality        = None



"""
    EnvInfo.py

    Copyright (c) 2019-2022, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.MachineTypes import get_bootup_state

env_info = None

def set_global_env_info(info):
    global env_info
    env_info = info

def get_global_env_info(gpu_info=False):
    global env_info
    if env_info is None:
        env_info = EnvInfo()
    if gpu_info:
        env_info.set_gpu_availability()
    return env_info

class EnvInfo:
    def __init__(self):
        bootup_state = get_bootup_state()
        self.normal = bootup_state.find("Normal") >= 0
        self.set_atsasl_availability()
        self.set_excel_availability()
        self.gpu_info = None               # self.set_gpu_availability() will be delayed
        self.latex_is_available = None

    def set_atsasl_availability(self):
        revoke_atsas = get_setting('revoke_atsas')
        if revoke_atsas:
            self.atsas_is_available = False
            return

        try:
            from molass_legacy.ATSAS.AtsasUtils import get_version
            self.atsas_version = get_version()
            self.atsas_is_available = self.atsas_version is not None
        except:
            # TODO: handle this better
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            print(etb)
            self.atsas_is_available = False

    def set_excel_availability(self):
        from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
        revoke_excel = get_setting('revoke_excel')
        if not self.normal or revoke_excel:
            self.excel_version = None
            self.excel_is_available = False
            self.com_error = None
            return

        try:
            from molass_legacy.KekLib.ExcelCOM import excel_availability
            self.excel_version = excel_availability()
            self.excel_is_available = self.excel_version is not None
            com_error = None
        except:
            com_error = str(ExceptionTracebacker())
            self.excel_is_available = False
        """
        set_setting('excel_is_available', self.excel_is_available)
            using persistent memory through SerialSettings to send info to the execution thread
            seems unreliable.
            
        """
        self.com_error = com_error

    def set_gpu_availability(self):
        revoke_cuda = get_setting('revoke_cuda')
        if revoke_cuda:
            self.cuda_tools_version = None
            self.nvidiagpu_is_available = False
            return

        if self.gpu_info is None:
            from .GpuInfo import GpuInfo
            self.gpu_info = gpu_info = GpuInfo()
            self.cuda_tools_version = gpu_info.cupy_version, gpu_info.cuda_version
            self.nvidiagpu_is_available = gpu_info.cupy_ok()
            if self.nvidiagpu_is_available:
                self.cuda_tools_ver_str = '(cupy %s, cuda %s)' % self.cuda_tools_version[0:2]
            else:
                self.cuda_tools_ver_str = ''

    def get_gpu_reason_texts(self):
        return self.gpu_info.get_reason_texts()

    def show_and_log_if_unavailable(self, dialog, logger):
        com_error = self.com_error
        if self.excel_is_available:
            ver_ = ' %s' % self.excel_version
            not_ = ''
        else:
            ver_ = ''
            not_ = 'not '
        excel_availability_message = 'Excel%s is %savailable.' % (ver_, not_)
        if com_error is not None:
            excel_availability_message += " com_error=" + com_error
        if self.excel_is_available:
            logger.info(excel_availability_message)
        else:
            logger.warning(excel_availability_message)
            if dialog is not None and get_setting('no_excel_warning'):
                from molass_legacy.SerialAnalyzer.NoExcelWarningDialog import NoExcelWarningDialog
                warn_dialog = NoExcelWarningDialog(dialog, com_error)
                warn_dialog.show()

        if self.atsas_is_available:
            ver_ = ' ' + self.atsas_version
            not_ = ''
        else:
            ver_ = '' 
            not_ = 'not '
        atsas_availability_message = 'ATSAS%s is %savailable.' % (ver_, not_)
        if self.atsas_is_available:
            logger.info(atsas_availability_message)
        else:
            logger.warning(atsas_availability_message)

        if self.gpu_info is not None:
            if self.nvidiagpu_is_available:
                cuda_ver = ' ' + self.cuda_tools_ver_str
                not_ = ''
            else:
                cuda_ver = ''
                not_ = 'not '
            nvidiagpu_availability_message = 'NVIDIA GPU%s is %savailable.' % (cuda_ver, not_)
            if self.nvidiagpu_is_available:
                logger.info(nvidiagpu_availability_message)
            else:
                logger.warning(nvidiagpu_availability_message)

    def get_latex_availability(self):
        if self.latex_is_available is None:
            try:
                import subprocess
                ret = subprocess.run(['latex', '-version'])
                self.latex_is_available = ret.returncode == 0
            except:
                self.latex_is_available = False
        return self.latex_is_available



"""
    Experiment.DataUtils.py

    Copyright (c) 2024-2025, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from molass_legacy.Experiment.ColumnTypes import get_columntype_from_id, get_columntype_from_name, PORESIZE_ALLOWANCE
from molass_legacy.SecTheory.MwRgFigure import get_mwrg_info

DEFAULT_NUM_PLATES = 48000

UNKNOWN_NAMES = [
    '20160227', 'HIF',
]

def get_columntype(in_folder=None):
    if in_folder is None:
        in_folder = get_setting('in_folder')

    test_columtype_id = get_setting('test_columtype_id')
    if test_columtype_id is not None:
        return get_columntype_from_id(test_columtype_id)

    columtype_id = None
    for column_name in UNKNOWN_NAMES:
        if in_folder.find(column_name) >= 0:
            columtype_id = 'unknown'
            return get_columntype_from_id(columtype_id)

    if in_folder.find('20230706') >= 0:
        columtype_id = 'agilentw'
    else:
        columtype_id = get_setting('default_columtype_id')

    return get_columntype_from_id(columtype_id)

def get_default_num_plates():
    columntype = get_columntype()
    num_plates_pm = columntype.num_plates
    if num_plates_pm is None or np.isnan(num_plates_pm):
        num_plates_pm = DEFAULT_NUM_PLATES
    return num_plates_pm

def update_sec_settings_impl(el_option, column_name, exclusion_limit, num_plates_pm):
    logger = logging.getLogger(__name__)

    if False:
        # currently suppressed due to the following error when an "Agilent" column is selected
        # UnicodeEncodeError: 'cp932' codec can't encode character '\xc5' in position 10964: illegal multibyte sequence
        logger.info("SEC column name is %s", column_name)     

    columntype, columntype_id = get_columntype_from_name(column_name, with_id=True)

    if columntype is None:
        exclusion_limit = get_setting('exclusion_limit')
        mwrg_info = get_mwrg_info()
        poresize = mwrg_info.compute_rg(exclusion_limit)
    else:
        poresize = columntype.poresize
        if el_option == 2:
            exclusion_limit = exclusion_limit
        else:
            exclusion_limit = columntype.excl_limit

    set_setting('columntype_id', columntype_id)
    set_setting('exclusion_limit', exclusion_limit)
    if poresize is None or np.isnan(poresize):
        unknown = 'unknown '
        poresize = 200
        set_setting('poresize_bounds', (35, 300))
    else:
        unknown = ''
        set_setting('poresize_bounds', (poresize - PORESIZE_ALLOWANCE, poresize + PORESIZE_ALLOWANCE))
    set_setting('poresize', poresize)

    if num_plates_pm is None or np.isnan(num_plates_pm):
        num_plates_pm = DEFAULT_NUM_PLATES
    num_plates_pc = int(round(num_plates_pm * 0.3))
    set_setting('num_plates_pc', num_plates_pc)

    logger.info("%spore size is set to %.1f", unknown, poresize)